#
# File: OpenMMUtil.py
# Author: Manish Sud <msud@san.rr.com>
#
# Copyright (C) 2025 Manish Sud. All rights reserved.
#
# The functionality available in this script is implemented using OpenMM, an
# open source molecuar simulation package.
#
# This file is part of MayaChemTools.
#
# MayaChemTools is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# MayaChemTools is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability of fitness
# for a particular purpose.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with MayaChemTools; if not, see <http://www.gnu.org/licenses/> or
# write to the Free Software Foundation Inc., 59 Temple Place, Suite 330,
# Boston, MA, 02111-1307, USA.
#

from __future__ import print_function

import os
import sys
import re
import glob
import multiprocessing as mp
import numpy as np

import openmm as mm
import openmm.app

import openmmforcefields as mmff
import openmmforcefields.generators
import openff as ff
import openff.toolkit

import pdbfixer
import mdtraj

import MiscUtil

__all__ = ["AddWaterBox", "DoAtomListsOverlap", "DoesAtomListOverlapWithSystemConstraints", "DoesSystemContainWater", "FreezeAtoms", "GetAtoms", "GetFormattedTotalSimulationTime", "InitializeBarostat", "InitializeIntegrator", "InitializeReporters", "InitializeSimulation", "InitializeSystem", "InitializeSystemGenerator", "MergeSmallMoleculeWithMacromolecule", "PerformAnnealing", "ProcessOptionOpenMMRestartParameters", "ProcessOptionOpenMMAtomsSelectionParameters", "ProcessOptionOpenMMAnnealingParameters", "ProcessOptionOpenMMForcefieldParameters", "ProcessOptionOpenMMIntegratorParameters", "ProcessOptionOpenMMPlatformParameters", "ProcessOptionOpenMMOutputParameters", "ProcessOptionOpenMMSimulationParameters", "ProcessOptionOpenMMSystemParameters", "ProcessOptionOpenMMWaterBoxParameters", "ReadPDBFile", "ReadSmallMoleculeFile", "RestraintAtoms", "SetupAnnealingParameters", "SetupIntegratorParameters", "SetupSimulationParameters", "SetupSystemGeneratorForcefieldsParameters", "WritePDBFile", "WriteSimulationStatePDBFile"]

def InitializeSystem(PDBFile, ForcefieldParamsInfo, SystemParamsInfo, WaterBoxAdd = False, WaterBoxParamsInfo = None, SmallMolFile = None, SmallMolID = "LIG"):
    """Initialize OpenMM system using specified forcefields, system and water box
    parameters along with a small molecule file and ID.
    
    The ForcefieldParamsInfo parameter is a dictionary of name and value pairs for
    system parameters and may be generated by calling the function named
    ProcessOptionOpenMMForcefieldParameters().
    
    The SystemParamsInfo parameter is a dictionary of name and value pairs for
    system parameters and may be generated by calling the function named
    ProcessOptionOpenMMSystemParameters().
    
    The WaterBoxParamsInfo parameter is a dictionary of name and value pairs for
    system parameters and may be generated by calling the function named
    ProcessOptionOpenMMWaterBoxParameters().

    Arguments:
        PDBFile (str): PDB file name..
        ForcefieldParamsInfo (dict): Parameter name and value pairs.
        SystemParamsInfo (dict): Parameter name and value pairs.
        WaterBoxAdd (bool): Add water box.
        WaterBoxParamsInfo (dict): Parameter name and value pairs.
        SmallMolFile (str): Small molecule file name..
        SmallMolID (str): Three letter small molecule ID

    Returns:
        Object: OpenMM system object.
        Object: OpenMM topology object.
        Object: OpenMM positions object.

    Examples:

        ... ... ...
        OptionsInfo["ForcefieldParams"] =
            OpenMMUtil.ProcessOptionOpenMMForcefieldParameters(
            "--forcefieldParams", Options["--forcefieldParams"])
        ... ... ...
        OptionsInfo["SystemParams"] = 
            OpenMMUtil.ProcessOptionOpenMMSystemParameters("--systemParams", 
            Options["--systemParams"])
        ... ... ...
        OptionsInfo["WaterBoxParams"] =
            OpenMMUtil.ProcessOptionOpenMMWaterBoxParameters(
            "--waterBoxParams", Options["--waterBoxParams"])
        ... ... ...
        System, Topology, Positions  = OpenMMUtil.InitializeSystem(
            OptionsInfo["Infile"], OptionsInfo["ForcefieldParams"],
            OptionsInfo["SystemParams"], OptionsInfo["WaterBox"],
            OptionsInfo["WaterBoxParams"], OptionsInfo["SmallMolFile"],
            OptionsInfo["SmallMolID"])

    """

    # Read PDB file file...
    MiscUtil.PrintInfo("\nReading PDB file %s..." % PDBFile)
    PDBHandle = ReadPDBFile(PDBFile)
    
    ModellerHandle = mm.app.Modeller(PDBHandle.topology, PDBHandle.positions)
    MiscUtil.PrintInfo("Number of residues: %s; Number of atoms: %s" % (ModellerHandle.topology.getNumResidues(), ModellerHandle.topology.getNumAtoms()))

    # Read small molecule file...
    SmallMols = None
    if SmallMolFile is not None:
        MiscUtil.PrintInfo("\nReading small molecule file %s..." % SmallMolFile)
        SmallMol = ReadSmallMoleculeFile(SmallMolFile)
        if SmallMol is None:
            MiscUtil.PrintError("Failed to read small molecule file: %s" % SmallMolFile)
        SmallMols = [SmallMol]

        MiscUtil.PrintInfo("\nGenerating macromolecule and small molecule complex...")
        MergeSmallMoleculeWithMacromolecule(ModellerHandle, SmallMol, SmallMolID)
        MiscUtil.PrintInfo("Number of residues: %s; Number of atoms: %s" % (ModellerHandle.topology.getNumResidues(), ModellerHandle.topology.getNumAtoms()))

    # Initialize system generator...
    BiopolymerForcefield = ForcefieldParamsInfo["Biopolymer"]
    SmallMoleculeForcefield = ForcefieldParamsInfo["SmallMolecule"]
    WaterForcefield = ForcefieldParamsInfo["Water"]
    AdditionalForcefiedsList = ForcefieldParamsInfo["AdditionalList"]
    SystemGeneratorHandle = InitializeSystemGenerator(BiopolymerForcefield, SmallMoleculeForcefield, WaterForcefield, SystemParamsInfo, SmallMols, AdditionalForcefiedsList)

    if WaterBoxAdd:
        AddWaterBox(ModellerHandle, SystemGeneratorHandle, WaterBoxParamsInfo)
        MiscUtil.PrintInfo("Number of residues: %s; Number of atoms: %s" % (ModellerHandle.topology.getNumResidues(), ModellerHandle.topology.getNumAtoms()))
    else:
        MiscUtil.PrintInfo("\nSkipping addition of a water box...")
        if DoesSystemContainWater(ModellerHandle.topology):
            if ForcefieldParamsInfo["ImplicitWater"]:
                MiscUtil.PrintInfo("Your system contains water molecules during the use of implicit water forcefield. The combination of biopolymer and water forcefields, %s and %s, specified using \"--forcefieldParams\" option may not be valid. You may consider removing water molecules from your system or specify a valid combination of biopolymer and water forcefields for explicit water." % (ForcefieldParamsInfo["Biopolymer"], ForcefieldParamsInfo["Water"]))

    MiscUtil.PrintInfo("\nBuilding system...")
    SystemHandle = SystemGeneratorHandle.create_system(ModellerHandle.topology, molecules = SmallMols)

    MiscUtil.PrintInfo("Periodic boundary conditions: %s" % ("Yes" if DoesSystemUsesPeriodicBoundaryConditions(SystemHandle) else "No"))

    return (SystemHandle, ModellerHandle.topology, ModellerHandle.positions)

def InitializeSystemGenerator(BiopolymerForcefield, SmallMoleculeForcefield, WaterForcefield, SystemParamsInfo, SmallMols, AdditionalForcefieldsList = None):
    """Initialize MMFF system generator using specified forcefields and system parameters
    along with a list of molecules.
    
    The SystemParamsInfo parameter is a dictionary of name and value pairs for
    system parameters and may be generated by calling the function named
    ProcessOptionOpenMMSystemParameters().

    Arguments:
        BiopolymerForcefield (str): Biopolymer force field name.
        SmallMoleculeForcefield (str): Small molecule force field name.
        WaterForcefield (str): Water force field name.
        SystemParamsInfo (dict): Parameter name and value pairs.
        SmallMols (list): List of OpenFF toolkit molecule objects.
        AdditionalForcefieldsList (list): List of any additional forcefield
            names

    Returns:
        Object: MMFF system generator object.

    Examples:

        OptionsInfo["SystemParams"] = 
            OpenMMUtil.ProcessOptionOpenMMSystemParameters("--systemParams", 
            Options["--systemParams"])
        ... ... ...
        SystemGeneratorHandle = OpenMMUtil.InitializeSystemGenerator(
            BiopolymerForcefield, SmallMoleculeForcefield, WaterForcefield,
            OptionsInfo["SystemParams"], SmallMols, AdditionalForcefiedsList)

    """

    (ForcefieldParams, PeriodicForcefieldParams, NonPeriodicForcefieldParams) = SetupSystemGeneratorForcefieldsParameters(SystemParamsInfo)

    AdditionalForcefieldMsg = ""
    if AdditionalForcefieldsList is not None:
        AdditionalForcefieldMsg = "; Additional forcefield(s): %s" % ", ".join(AdditionalForcefieldsList)
        
    MiscUtil.PrintInfo("\nInitializing system generator (Biopolymer forcefield: %s; Small molecule forcefield: %s; Water forcefield: %s%s)..." % (BiopolymerForcefield, SmallMoleculeForcefield,WaterForcefield, AdditionalForcefieldMsg))

    ForcefieldsList = [BiopolymerForcefield, WaterForcefield]
    if AdditionalForcefieldsList is not None:
        ForcefieldsList.extend(AdditionalForcefieldsList)

    SystemGeneratorHandle = mmff.generators.SystemGenerator(forcefields = ForcefieldsList, small_molecule_forcefield = SmallMoleculeForcefield, molecules = SmallMols, forcefield_kwargs = ForcefieldParams, periodic_forcefield_kwargs = PeriodicForcefieldParams, nonperiodic_forcefield_kwargs = NonPeriodicForcefieldParams)

    return SystemGeneratorHandle

def InitializeIntegrator(ParamsInfo, ConstraintErrorTolerance):
    """Initialize integrator.
    
    The ParamsInfo parameter is a dictionary of name and value pairs for
    integrator parameters and may be generated by calling the function named
    ProcessOptionOpenMMIntegratorParameters().

    Arguments:
        ParamsInfo (dict): Parameter name and value pairs.
        ConstraintErrorTolerance (float): Distance tolerance for
            constraints as a fraction of the constrained distance.

    Returns:
        Object: OpenMM integrator object.

    Examples:

        OptionsInfo["IntegratorParams"] =
            OpenMMUtil.ProcessOptionOpenMMIntegratorParameters(
            "--integratorParams", Options["--integratorParams"],
            HydrogenMassRepartioningStatus =
            OptionsInfo["SystemParams"]["HydrogenMassRepartioning"])
        ... ... ...
        Integrator = OpenMMUtil.InitializeIntegrator(
            OptionsInfo["IntegratorParams"],
            OptionsInfo["SystemParams"]["ConstraintErrorTolerance"])

    """

    IntegratorParamsInfo = SetupIntegratorParameters(ParamsInfo)

    IntegratorName = IntegratorParamsInfo["Integrator"]
    RandomSeed = IntegratorParamsInfo["RandomSeed"]
    StepSize = IntegratorParamsInfo["StepSize"]
    Temperature = IntegratorParamsInfo["Temperature"]
    FrictionCoefficient = IntegratorParamsInfo["FrictionCoefficient"]
    
    MiscUtil.PrintInfo("\nIntializing integrator (Name: %s; StepSize: %s; Temperature: %s)..." % (IntegratorName, StepSize, Temperature))

    if re.match("^LangevinMiddle$", IntegratorName, re.I):
        Integrator = mm.LangevinMiddleIntegrator(Temperature, FrictionCoefficient, StepSize)
    elif re.match("^Langevin$", IntegratorName, re.I):
        Integrator = mm.LangevinIntegrator(Temperature, FrictionCoefficient, StepSize)
    elif re.match("^NoseHoover$", IntegratorName, re.I):
        Integrator = mm.NoseHooverIntegrator(Temperature, FrictionCoefficient, StepSize)
    elif re.match("^Brownian$", IntegratorName, re.I):
        Integrator = mm.BrownianIntegrator(Temperature, FrictionCoefficient, StepSize)
    else:
        MiscUtil.PrintError("The parameter value specified, %s, for parameter name, integrator, for option \"--integratorParams\" is not a valid value. Supported values: LangevinMiddle, Langevin, NoseHoover, or Brownian" % IntegratorName)

    Integrator.setConstraintTolerance(ConstraintErrorTolerance)

    if RandomSeed is not None:
        if re.match("^(LangevinMiddle|Langevin|Brownian)$", IntegratorName, re.I):
            MiscUtil.PrintInfo("Setting random number seed for integrator to %s..." % RandomSeed)
            Integrator.setRandomNumberSeed(RandomSeed)
        else:
            MiscUtil.PrintInfo("Skipping setting of random number seed. Not supported for integrator %s..." % IntegratorName)

    return Integrator

def InitializeBarostat(ParamsInfo):
    """Initialize barostat.
    
    The ParamsInfo parameter is a dictionary of name and value pairs for
    integrator parameters and may be generated by calling the function named
    ProcessOptionOpenMMIntegratorParameters().

    Arguments:
        ParamsInfo (dict): Parameter name and value pairs.

    Returns:
        Object: OpenMM barostat object.

    Examples:

        OptionsInfo["IntegratorParams"] =
            OpenMMUtil.ProcessOptionOpenMMIntegratorParameters(
            "--integratorParams", Options["--integratorParams"],
            HydrogenMassRepartioningStatus =
            OptionsInfo["SystemParams"]["HydrogenMassRepartioning"])
        ... ... ...
        Barostat = OpenMMUtil.InitializeBarostat(
            OptionsInfo["IntegratorParams"])

    """
    IntegratorParamsInfo = SetupIntegratorParameters(ParamsInfo)

    BarostatName = IntegratorParamsInfo["Barostat"]
    if re.match("^MonteCarlo$", BarostatName, re.I):
        MiscUtil.PrintInfo("\nInitializing Monte Carlo barostat (Pressure: %s)... " % (IntegratorParamsInfo["Pressure"]))
        Barostat = mm.MonteCarloBarostat(IntegratorParamsInfo["Pressure"], IntegratorParamsInfo["Temperature"], IntegratorParamsInfo["BarostatInterval"])
    elif re.match("^MonteCarloMembrane$", BarostatName, re.I):
        MiscUtil.PrintInfo("\nInitializing Monte Carlo membrane barostat (Pressure: %s; SurfaceTension: %s; XYMode: %s; ZMode: %s)... " % (IntegratorParamsInfo["Pressure"], IntegratorParamsInfo["SurfaceTension"], IntegratorParamsInfo["XYModeSpecified"], IntegratorParamsInfo["ZModeSpecified"]))
        Barostat = mm.MonteCarloMembraneBarostat(IntegratorParamsInfo["Pressure"], IntegratorParamsInfo["SurfaceTension"], IntegratorParamsInfo["Temperature"], IntegratorParamsInfo["XYMode"], IntegratorParamsInfo["ZMode"], IntegratorParamsInfo["BarostatInterval"])
    else:
        MiscUtil.PrintError("The parameter value specified, %s, for parameter name, barostat, for option \"--integratorParams\" is not a valid value. Supported values: MonteCarlo or MonteCarloMembrane" % BarostatName)

    if IntegratorParamsInfo["RandomSeed"] is not None:
        RandomSeed = IntegratorParamsInfo["RandomSeed"]
        MiscUtil.PrintInfo("Setting random number seed for barostat to %s..." % RandomSeed)
        Barostat.setRandomNumberSeed(RandomSeed)

    return Barostat

def InitializeSimulation(System, Integrator, Topology, Positions, PlatformParamsInfo):
    """Initialize simulation.
    
    The PlatformParamsInfo parameter is a dictionary of name and value pairs for
    platform parameters and may be generated by calling the function named
    ProcessOptionOpenMMPlatformParameters().

    Arguments:
        System (object): OpenMM system object.
        Integrator (object): OpenMM integrator object.
        Topology (object): OpenMM topology object.
        Positons (object): OpenMM Positions object.
        PlatformParamsInfo (dict): Parameter name and value pairs.

    Returns:
        Object: OpenMM simulation object.

    Examples:

        ParamsDefaultInfoOverride = {"Name": Options["--platform"],
             "Threads": 1}
        OptionsInfo["PlatformParams"] =
            OpenMMUtil.ProcessOptionOpenMMPlatformParameters("--platformParams",
            Options["--platformParams"], ParamsDefaultInfoOverride)
        ... ... ...
        Simulation = OpenMMUtil.InitializeSimulation(System, Integrator, Topology,
            Positions, OptionsInfo["PlatformParams"])

    """

    PlatformName, PlatformProperties, PlatformMsg = SetupPlatformParameters(PlatformParamsInfo)
    MiscUtil.PrintInfo("\nInitializing simulation (%s)..." % PlatformMsg)

    try:
        PlatformHandle = mm.Platform.getPlatformByName(PlatformName)
    except Exception as ErrMsg:
        MiscUtil.PrintInfo("")
        MiscUtil.PrintError("Failed to get platform %s:\n%s\n" % (PlatformName, ErrMsg))
        
    try:
        SimulationHandle = mm.app.Simulation(Topology, System, Integrator, PlatformHandle, PlatformProperties)
    except Exception as ErrMsg:
        MiscUtil.PrintInfo("")
        MiscUtil.PrintError("Failed to initialize simulation: %s\n" % (ErrMsg))

    SimulationHandle.context.setPositions(Positions)

    return (SimulationHandle)

def InitializeReporters(OutputParamsInfo, TotalSteps, DataOutAppendStatus):
    """Initialize reporters for writing data to trajectory, log, and checkpoint
    files along with reporting to the stdout.
    
    The OutputParamsInfo parameter is a dictionary of name and value pairs for
    output parameters and may be generated by calling the function named
    ProcessOptionOpenMMOutputParameters().

    Arguments:
        OutputParamsInfo (dict): Parameter name and value pairs.
        TotalSteps (int): Total number of simulation steps.
        DataOutAppendStatus (bool): Append data to trajectory and log file.

    Returns:
        Object: OpenMM trajectory reporter object.
        Object: OpenMM data log file reporter object.
        Object: OpenMM stodut reporter object.
        Object: OpenMM checkpoint reporter object.

    Examples:

        if OptionsInfo["NVTMode"]:
            ParamsDefaultInfoOverride = {"DataOutType": "Step Speed Progress
                PotentialEnergy Temperature Time Volume"}
        else:
            ParamsDefaultInfoOverride = {"DataOutType": "Step Speed Progress
              PotentialEnergy Temperature Time Density"}
        OptionsInfo["OutputParams"] =
            OpenMMUtil.ProcessOptionOpenMMOutputParameters("--outputParams", 
            Options["--outputParams"], OptionsInfo["OutfilePrefix"],
            ParamsDefaultInfoOverride)
        ProcessOutfileNames()
        ... ... ...
        (TrajReporter, DataLogReporter, DataStdoutReporter, CheckpointReporter)
            = OpenMMUtil.InitializeReporters(OptionsInfo["OutputParams"],
            OptionsInfo["SimulationParams"]["Steps"], OptionsInfo["DataOutAppendMode"])

    """

    (TrajReporter, DataLogReporter, DataStdoutReporter, CheckpointReporter) = [None] * 4
    if OutputParamsInfo["Traj"]:
        if re.match("^DCD$", OutputParamsInfo["TrajFormat"], re.I):
            TrajReporter = mm.app.DCDReporter(OutputParamsInfo["TrajFile"], OutputParamsInfo["TrajSteps"], append = DataOutAppendStatus)
        elif re.match("^XTC$", OutputParamsInfo["TrajFormat"], re.I):
            if not DataOutAppendStatus:
                # Remove existing traj file; otherwise, XTCReporter fails...
                if os.path.isfile(OutputParamsInfo["TrajFile"]):
                    os.remove(OutputParamsInfo["TrajFile"])
            TrajReporter = mm.app.XTCReporter(OutputParamsInfo["TrajFile"], OutputParamsInfo["TrajSteps"], append = DataOutAppendStatus)
        else:
            MiscUtil.PrintError("The parameter value specified, %s, for parameter name, trajFormat, for option \"--outputParams\" is not a valid value. Supported format: DCD or XTC")

    if OutputParamsInfo["Checkpoint"]:
        CheckpointReporter = mm.app.CheckpointReporter(OutputParamsInfo["CheckpointFile"], OutputParamsInfo["CheckpointSteps"])

    DataOutTypeStatusMap = OutputParamsInfo["DataOutTypeStatusMap"]
    if OutputParamsInfo["DataLog"]:
        DataLogReporter = mm.app.StateDataReporter(OutputParamsInfo["DataLogFile"], OutputParamsInfo["DataLogSteps"], totalSteps = TotalSteps, step = DataOutTypeStatusMap["Step"], time = DataOutTypeStatusMap["Time"], speed = DataOutTypeStatusMap["Speed"], progress = DataOutTypeStatusMap["Progress"], elapsedTime = DataOutTypeStatusMap["ElapsedTime"], remainingTime = DataOutTypeStatusMap["RemainingTime"], potentialEnergy = DataOutTypeStatusMap["PotentialEnergy"], kineticEnergy = DataOutTypeStatusMap["KineticEnergy"], totalEnergy = DataOutTypeStatusMap["TotalEnergy"], temperature = DataOutTypeStatusMap["Temperature"], volume = DataOutTypeStatusMap["Volume"], density = DataOutTypeStatusMap["Density"], separator = OutputParamsInfo["DataOutDelimiter"], append = DataOutAppendStatus)
        
    if OutputParamsInfo["DataStdout"]:
        DataStdoutReporter = mm.app.StateDataReporter(sys.stdout, OutputParamsInfo["DataStdoutSteps"], totalSteps = TotalSteps, step = DataOutTypeStatusMap["Step"], time = DataOutTypeStatusMap["Time"], speed = DataOutTypeStatusMap["Speed"], progress = DataOutTypeStatusMap["Progress"], elapsedTime = DataOutTypeStatusMap["ElapsedTime"], remainingTime = DataOutTypeStatusMap["RemainingTime"], potentialEnergy = DataOutTypeStatusMap["PotentialEnergy"], kineticEnergy = DataOutTypeStatusMap["KineticEnergy"], totalEnergy = DataOutTypeStatusMap["TotalEnergy"], temperature = DataOutTypeStatusMap["Temperature"], volume = DataOutTypeStatusMap["Volume"], density = DataOutTypeStatusMap["Density"], separator = OutputParamsInfo["DataOutDelimiter"])
        
    return (TrajReporter, DataLogReporter, DataStdoutReporter, CheckpointReporter)

def PerformAnnealing(Simulation, Integrator, Barostat, TemperatureStart, TemperatureEnd, TemperatureChange, SimulationSteps):
    """Perform annealing by heating or cooling the system from start to end
    temperature.
    
    The temperature is increased or decreased from start to end temperature by
    temperature to perform annealing following by performing simulations for a
    specified number of steps after each temperature step.

    Arguments:
        System (object): OpenMM system object.
        Integrator (object): OpenMM integrator object.
        Barostat (object): OpenMM integrator object.
        TemperatureStart (float): Start temperature.
        TemperatureEnd (float): End temperature.
        TemperatureChange (int): Temperature step size to heat or cool the
            system from start to end temperature.
        SimulationSteps (fint): Number of simulations steps to perform after
            each temperature step.

    Returns:
        int: Total number of simulation steps.

    Examples:

        ... ... ...
        TotalInitialSimulationSteps  = OpenMMUtil.PerformAnnealing(Simulation,
            Integrator, Barostat, InitialStart, InitialEnd, InitialChange, InitialSteps)

    """

    if TemperatureStart < TemperatureEnd:
        TemperatureRange = np.arange(TemperatureStart, TemperatureEnd, TemperatureChange)
    else:
        TemperatureRange = np.arange(TemperatureStart, TemperatureEnd, -TemperatureChange)

    TemperatureRange = TemperatureRange.tolist()
    TemperatureRange.append(TemperatureEnd)

    TotalSimulationSteps = 0
    for Temperature in TemperatureRange:
        Temperature = Temperature * mm.unit.kelvin

        Integrator.setTemperature(Temperature)
        if Barostat is not None:
            Barostat.setDefaultTemperature(Temperature)

        Simulation.step(SimulationSteps)
        TotalSimulationSteps += SimulationSteps

    return TotalSimulationSteps

def AddWaterBox(ModellerHandle, SystemGeneratorHandle, WaterBoxParamsInfo):
    """Add a water box.
    
    The WaterBoxParamsInfo parameter is a dictionary of name and value pairs for
    waterbox parameters and may be generated by calling the function named
    ProcessOptionOpenMMWaterBoxParameters().

    Arguments:
        ModellerHandle (object): OpenMM modeller object.
        SystemGeneratorHandle (object): OpenMM system generator object.
        WaterBoxParamsInfo (dict): Parameter name and value pairs.

    Returns:
        None.

    Examples:

        OptionsInfo["WaterBoxParams"] =
            OpenMMUtil.ProcessOptionOpenMMWaterBoxParameters("--waterBoxParams",
            Options["--waterBoxParams"])
        ... ... ...
        OpenMMUtil.AddWaterBox(ModellerHandle, SystemGeneratorHandle,
           OptionsInfo["WaterBoxParams"])

    """

    MiscUtil.PrintInfo("\nRemoving any existing waters...")
    ModellerHandle.deleteWater()

    MiscUtil.PrintInfo("\nAdding a water box...")
    
    Size, Padding, Shape = [None] * 3
    if WaterBoxParamsInfo["ModeSize"]:
        SizeList = WaterBoxParamsInfo["SizeList"]
        Size = mm.Vec3(SizeList[0], SizeList[1], SizeList[2]) * mm.unit.nanometer
    elif WaterBoxParamsInfo["ModePadding"]:
        Padding = WaterBoxParamsInfo["Padding"] * mm.unit.nanometer
        Shape = WaterBoxParamsInfo["Shape"]
    else:
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, mode, using \"--waterBoxParams\" option is not a valid value. Supported values: Size Padding \n" % (WaterBoxParamsInfo["Mode"]))

    IonicStrength = WaterBoxParamsInfo["IonicStrength"] * mm.unit.molar

    ModellerHandle.addSolvent(SystemGeneratorHandle.forcefield, model = WaterBoxParamsInfo["Model"], boxSize = Size, boxVectors = None, padding = Padding, numAdded = None, boxShape = Shape, positiveIon = WaterBoxParamsInfo["IonPositive"], negativeIon = WaterBoxParamsInfo["IonNegative"], ionicStrength = IonicStrength, neutralize = True)

def SetupSystemGeneratorForcefieldsParameters(SystemParamsInfo):
    """Setup forcefied parameters for OpenMM API calls.
    
    The SystemParamsInfo parameter is a dictionary of name and value pairs for
    system parameters and may be generated by calling the function named
    ProcessOptionOpenMMSystemParameters().

    Arguments:
        SystemParamsInfo (dict): Parameter name and value pairs.

    Returns:
        dict: Forcefield parameter name and value pairs.
        dictionary2: Periodic forcefield parameter name and value pairs.
        dictionary3: Non-periodic parameter name and value pairs.

    Examples:

        OptionsInfo["SystemParams"] = 
            OpenMMUtil.ProcessOptionOpenMMSystemParameters("--systemParams", 
            Options["--systemParams"])
        ... ... ...
        (ForcefieldParams, PeriodicForcefieldParams, NonPeriodicForcefieldParams)
            = SetupSystemGeneratorForcefieldsParameters(OptionsInfo["SystemParams"])

    """

    ForcefieldParams = {"constraints": SystemParamsInfo["Constraints"], "rigidWater": SystemParamsInfo["RigidWater"], "removeCMMotion": SystemParamsInfo["RemoveCMMotion"]}
    if SystemParamsInfo["HydrogenMassRepartioning"]:
        ForcefieldParams["hydrogenMass"] = SystemParamsInfo["HydrogenMass"] * mm.unit.amu

    NonbondedCutoff = SystemParamsInfo["NonbondedCutoff"] * mm.unit.nanometers
    PeriodicForcefieldParams = {"nonbondedMethod": SystemParamsInfo["NonbondedMethodPeriodic"], "nonbondedCutoff": NonbondedCutoff, "ewaldErrorTolerance":  SystemParamsInfo["EwaldErrorTolerance"]}
    NonPeriodicForcefieldParams = {"nonbondedMethod": SystemParamsInfo["NonbondedMethodNonPeriodic"], "nonbondedCutoff" : NonbondedCutoff}

    return (ForcefieldParams, PeriodicForcefieldParams, NonPeriodicForcefieldParams)

def SetupPlatformParameters(ParamsInfo):
    """Setup platform parameters for OpenMM calls.
    
    The ParamsInfo parameter is a dictionary of name and value pairs for
    platform parameters and may be generated by calling the function named
    ProcessOptionOpenMMPlatformParameters().

    Arguments:
        ParamsInfo (dict): Parameter name and value pairs.

    Returns:
        str: PlatformName.
        dict: Platform properities parameter name and values pairs.
        str: Text message describing platform.

    Examples:

        ParamsDefaultInfoOverride = {"Name": Options["--platform"],
             "Threads": 1}
        OptionsInfo["PlatformParams"] =
            OpenMMUtil.ProcessOptionOpenMMPlatformParameters("--platformParams",
            Options["--platformParams"], ParamsDefaultInfoOverride)
        ... ... ...
        PlatformName, PlatformProperties, PlatformMsg =
            SetupPlatformParameters(PlatformParamsInfo)

    """

    PlatformName = ParamsInfo["Name"]
    ParamNames = None

    if re.match("^CPU$", PlatformName, re.I):
        ParamNames = ["Threads"]
    elif re.match("^CUDA$", PlatformName, re.I):
        ParamNames = ["DeviceIndex", "DeterministicForces", "Precision", "TempDirectory", "UseBlockingSync", "UseCpuPme"]
    elif re.match("^OpenCL$", PlatformName, re.I):
        ParamNames = ["DeviceIndex", "OpenCLPlatformIndex", "Precision", "UseCpuPme"]
    elif re.match("^Reference$", PlatformName, re.I):
        ParamNames = None
    else:
        MiscUtil.PrintError("The parameter value specified, %s, for parameter name, name, for option \"--platformParams\" is not a valid value. Supported values: CPU, CUDA, OpenCL, or Reference")

    PlatformProperties = None
    if ParamNames is not None:
        FirstValue = True
        for ParamName in ParamNames:
            ParamValue = ParamsInfo[ParamName]
            if ParamValue is not None:
                if FirstValue:
                    FirstValue = False
                    PlatformProperties = {}
                PlatformProperties[ParamName] = ParamValue

    PlatformMsg = "Platform: %s" % PlatformName
    if re.match("^CPU$", PlatformName, re.I):
        Threads = ParamsInfo["Threads"]
        if Threads is None or Threads == "0":
            Threads = "auto"
        PlatformMsg = "Platform: %s; Threads: %s" % (PlatformName, Threads)
    elif re.match("^(CUDA|OpenCL)$", PlatformName, re.I):
        DeviceIndex = "auto" if ParamsInfo["DeviceIndex"] is None else ParamsInfo["DeviceIndex"]
        Precision = "auto" if ParamsInfo["Precision"] is None else ParamsInfo["Precision"]
        PlatformMsg = "Platform: %s; DeviceIndex: %s; Precision: %s" % (PlatformName, DeviceIndex, Precision)

    return (PlatformName, PlatformProperties, PlatformMsg)

def SetupAnnealingParameters(ParamsInfo):
    """Setup annealing parameters for OpenMM API calls.
    
    The ParamsInfo parameter is a dictionary of name and value pairs for
    annealing parameters and may be generated by calling the function named
    ProcessOptionOpenMMAnnealingParameters().

    Arguments:
        ParamsInfo (dict): Parameter name and value pairs.

    Returns:
        dict: Integrator parameter name and values pairs.

    Examples:

        OptionsInfo["AnnealingParams"] =
            OpenMMUtil.ProcessOptionOpenMMAnnealingParameters(
            "--annealingParams", Options["--annealingParams"],
        ... ... ...
        IntegratorParams = SetupIntegratorParameters(
            OptionsInfo["IntegratorParams"])

    """

    ParamsInfoWithUnits = {}
    
    ParamsInfoWithUnits["InitialStart"] = ParamsInfo["InitialStart"] * mm.unit.kelvin
    ParamsInfoWithUnits["InitialEnd"] = ParamsInfo["InitialEnd"] * mm.unit.kelvin
    ParamsInfoWithUnits["InitialChange"] = ParamsInfo["InitialChange"] * mm.unit.kelvin
    ParamsInfoWithUnits["InitialSteps"] = ParamsInfo["InitialSteps"]

    ParamsInfoWithUnits["InitialEquilibrationSteps"] = ParamsInfo["InitialEquilibrationSteps"]

    ParamsInfoWithUnits["Cycles"] = ParamsInfo["Cycles"]
    ParamsInfoWithUnits["CycleStart"] = ParamsInfo["CycleStart"] * mm.unit.kelvin
    ParamsInfoWithUnits["CycleEnd"] = ParamsInfo["CycleEnd"] * mm.unit.kelvin
    ParamsInfoWithUnits["CycleChange"] = ParamsInfo["CycleChange"] * mm.unit.kelvin
    ParamsInfoWithUnits["CycleSteps"] = ParamsInfo["CycleSteps"]

    ParamsInfoWithUnits["CycleEquilibrationSteps"] = ParamsInfo["CycleEquilibrationSteps"]

    ParamsInfoWithUnits["FinalEquilibrationSteps"] = ParamsInfo["FinalEquilibrationSteps"]

    return ParamsInfoWithUnits

def SetupIntegratorParameters(ParamsInfo):
    """Setup integrator parameters for OpenMM API calls.
    
    The ParamsInfo parameter is a dictionary of name and value pairs for
    integrator parameters and may be generated by calling the function named
    ProcessOptionOpenMMIntegratorParameters().

    Arguments:
        ParamsInfo (dict): Parameter name and value pairs.

    Returns:
        dict: Integrator parameter name and values pairs.

    Examples:

        OptionsInfo["IntegratorParams"] =
            OpenMMUtil.ProcessOptionOpenMMIntegratorParameters(
            "--integratorParams", Options["--integratorParams"],
            HydrogenMassRepartioningStatus =
            OptionsInfo["SystemParams"]["HydrogenMassRepartioning"])
        ... ... ...
        IntegratorParams = SetupIntegratorParameters(
            OptionsInfo["IntegratorParams"])

    """

    ParamsInfoWithUnits = {}
    
    ParamsInfoWithUnits["Integrator"] = ParamsInfo["Integrator"]

    ParamsInfoWithUnits["RandomSeed"] = ParamsInfo["RandomSeed"]

    ParamsInfoWithUnits["FrictionCoefficient"] = ParamsInfo["FrictionCoefficient"]/mm.unit.picosecond
    ParamsInfoWithUnits["StepSize"] = ParamsInfo["StepSize"] * mm.unit.femtoseconds
    ParamsInfoWithUnits["Temperature"] = ParamsInfo["Temperature"] * mm.unit.kelvin

    ParamsInfoWithUnits["Barostat"] = ParamsInfo["Barostat"]

    ParamsInfoWithUnits["Pressure"] = ParamsInfo["Pressure"] * mm.unit.atmospheres
    ParamsInfoWithUnits["BarostatInterval"] = ParamsInfo["BarostatInterval"]

    SurfaceTensionInAngstroms = ParamsInfo["SurfaceTension"]
    SurfaceTension = SurfaceTensionInAngstroms  * mm.unit.atmospheres * mm.unit.angstroms

    SurfaceTensionInNanometers = SurfaceTension.value_in_unit(mm.unit.atmospheres * mm.unit.nanometer)
    ParamsInfoWithUnits["SurfaceTension"] = SurfaceTensionInNanometers * mm.unit.atmospheres * mm.unit.nanometers

    XYMode = ParamsInfo["XYMode"]
    XYModeSpecified = ""
    if re.match("^Anisotropic$", XYMode, re.I):
        XYMode = mm.MonteCarloMembraneBarostat.XYAnisotropic
        XYModeSpecified = "Anisotropic"
    elif re.match("^Isotropic$", XYMode, re.I):
        XYMode = mm.MonteCarloMembraneBarostat.XYIsotropic
        XYModeSpecified = "Isotropic"
    else:
        MiscUtil.PrintError("The parameter value specified, %s, for parameter name, xymode, for option \"--integratorParams\" is not a valid value. Supported values: Anisotropic or  Isotropic" % XYMode)
    ParamsInfoWithUnits["XYMode"] = XYMode
    ParamsInfoWithUnits["XYModeSpecified"] = XYModeSpecified

    ZMode = ParamsInfo["ZMode"]
    ZModeSpecified = ""
    if re.match("^Fixed$", ZMode, re.I):
        ZMode = mm.MonteCarloMembraneBarostat.ZFixed
        ZModeSpecified = "Fixed"
    elif re.match("^Free$", ZMode, re.I):
        ZMode = mm.MonteCarloMembraneBarostat.ZFree
        ZModeSpecified = "Free"
    else:
        MiscUtil.PrintError("The parameter value specified, %s, for parameter name, zmode, for option \"--integratorParams\" is not a valid value. Supported values: Fixed or Free" % ZMode)
    ParamsInfoWithUnits["ZMode"] = ZMode
    ParamsInfoWithUnits["ZModeSpecified"] = ZModeSpecified

    return ParamsInfoWithUnits

def SetupSimulationParameters(ParamsInfo):
    """Setup simulation parameters for OpenMM API calls.
    
    The ParamsInfo parameter is a dictionary of name and value pairs for
    integrator parameters and may be generated by calling the function named
    ProcessOptionOpenMSimulationParameters().

    Arguments:
        ParamsInfo (dict): Parameter name and value pairs.

    Returns:
        dict: Integrator parameter name and values pairs.

    Examples:

        OptionsInfo["SimulationParams"] =
            OpenMMUtil.ProcessOptionOpenMMSimulationParameters(
            "--simulationParams", Options["--simulationParams"])
        ... ... ...
        SimulationParams = SetupSimulationParameters(
            OptionsInfo["SimulationParams"])

    """

    ParamsInfoWithUnits = {}

    ParamsInfoWithUnits["Steps"] = ParamsInfo["Steps"]

    ParamsInfoWithUnits["Minimization"] = ParamsInfo["Minimization"]
    ParamsInfoWithUnits["MinimizationMaxSteps"] = ParamsInfo["MinimizationMaxSteps"]

    MinimizationToleranceInKcal = ParamsInfo["MinimizationTolerance"]
    MinimizationTolerance = MinimizationToleranceInKcal * mm.unit.kilocalories_per_mole/mm.unit.angstroms
    
    MinimizationToleranceInJoules = MinimizationTolerance.value_in_unit(mm.unit.kilojoules_per_mole/mm.unit.nanometer)
    MinimizationTolerance = MinimizationToleranceInJoules*mm.unit.kilojoules_per_mole/mm.unit.nanometer

    ParamsInfoWithUnits["MinimizationToleranceInKcal"] = MinimizationToleranceInKcal
    ParamsInfoWithUnits["MinimizationToleranceInJoules"] = MinimizationToleranceInJoules
    ParamsInfoWithUnits["MinimizationTolerance"] = MinimizationTolerance
    
    ParamsInfoWithUnits["Equilibration"] = ParamsInfo["Equilibration"]
    ParamsInfoWithUnits["EquilibrationSteps"] = ParamsInfo["EquilibrationSteps"]

    return ParamsInfoWithUnits
    
def ReimageRealignTrajectory(Topology, TrajFile, ReimageFrames = True, RealignFrames = True, Selection = "protein and backbone and name CA"):
    """Reimage and realign a trajectory file using MDTraj. The trajectory frames
    are reimaged before realigning to the first frame using the specified atom
    selection.
    
    The trajectory file format must a valid format supported by MDTraj. No
    validation is performed.

    Arguments:
        Topology (str or object): PDB file name or OpenMM topology object.
        TrajFile (str): Trajectory file name.
        ReimageFrames (bool): Reimage trajectory frames.
        RealignFrames (bool): Realign trajectory frames.
        Selection (str): MDTraj atom selection for realigning frames.

    Returns:
        None or object: MDTraj trajectory object.
        bool: Reimaged status.
        bool: Realigned status.

    """

    if isinstance(Topology, str):
        MiscUtil.PrintInfo("Reading trajectory file %s (TopologyFile: %s)..." % (TrajFile, Topology))
    else:
        MiscUtil.PrintInfo("Reading trajectory file %s..." % (TrajFile))
        Topology = mdtraj.Topology.from_openmm(Topology)

    try:
        Traj = mdtraj.load(TrajFile, top = Topology)
    except Exception as ErrMsg:
        MiscUtil.PrintInfo("")
        MiscUtil.PrintWarning("Failed to read trajectory file: %s" % ErrMsg)
        MiscUtil.PrintInfo("")
        return (None, False, False)

    ReimagedStatus, RealignedStatus = [False] * 2

    if ReimageFrames:
        MiscUtil.PrintInfo("Reimaging frames...")
        try:
            Traj.image_molecules(inplace = True)
            ReimagedStatus = True
        except Exception as ErrMsg:
            MiscUtil.PrintInfo("")
            MiscUtil.PrintWarning("Failed to reimage frames: %s" % ErrMsg)
            MiscUtil.PrintInfo("")
    else:
        MiscUtil.PrintInfo("Skipping reimaging of frames...")

    if RealignFrames:
        MiscUtil.PrintInfo("Realigning frames to the first frame using selection \"%s\"..." % Selection)
        try:
            SelectionAtomIndices = Traj.top.select(Selection)
        except Exception as ErrMsg:
            MiscUtil.PrintInfo("")
            MiscUtil.PrintWarning("Failed to align frames using selection \"%s\": %s" % (Selection, ErrMsg))
            MiscUtil.PrintInfo("")
            return (Traj, ReimagedStatus, RealignedStatus)

        if SelectionAtomIndices.size == 0:
            MiscUtil.PrintInfo("")
            MiscUtil.PrintWarning("Failed to align frames using selection \"%s\": No matched atoms found" % (Selection))
            MiscUtil.PrintInfo("")
            return (Traj, ReimagedStatus, RealignedStatus)
            
        RealignedStatus = True
        Traj.superpose(Traj, frame = 0, atom_indices = SelectionAtomIndices)
    else:
        MiscUtil.PrintInfo("Skipping realignment of frames to the first frame...")

    return (Traj, ReimagedStatus, RealignedStatus)

def FreezeAtoms(System, AtomList):
    """Freeze atoms during a simulation. The specified atoms are kept completely
    fixed by setting their masses to zero. Their positions do not change during
    local energy minimization and MD simulation, and they do not contribute
    to the kinetic energy of the system.

    Arguments:
        System (object): OpenMM system object.
        AtomList (list): List of OpenMM atom objects.

    Returns:
        None

    """

    if AtomList is None:
        return

    for Atom in AtomList:
        System.setParticleMass(Atom.index, 0*mm.unit.amu)

def RestraintAtoms(System, Positions, AtomList, SpringConstantInKcal):
    """Restraint atoms during a simulation. The motion of specified atoms is
        restricted by adding a harmonic force that binds them to their starting
        positions. The atoms are not completely fixed unlike freezing of atoms.
        Their motion, however, is restricted and they are not able to move far away
        from their starting positions during local energy minimization and MD
        simulation.
        
        The SpringConstantInKcal value must be specified in the units of
        kcal/mol/A*82. It is automatically converted into the units of 
        kjoules/mol/nm**2 for OpenMM API call.

    Arguments:
        System (object): OpenMM system object.
        Positions (object): OpenMM positons object object.
        AtomList (list): List of OpenMM atom object.
        SpringConstantInKcal (float): Spring constant value.

    Returns:
        None

    """

    if AtomList is None:
        return

    SpringConstant =  SpringConstantInKcal * mm.unit.kilocalories_per_mole/mm.unit.angstroms**2
    SpringConstantInKjoules = SpringConstant.value_in_unit(mm.unit.kilojoules_per_mole/mm.unit.nanometer**2)

    MiscUtil.PrintInfo("Restraint spring constant: %.2f kcal/mol/A**2 (%.2f kjoules/mol/nm**2)" % (SpringConstantInKcal, SpringConstantInKjoules))

    if DoesSystemUsesPeriodicBoundaryConditions(System):
        # For periodic systems...
        RestraintForce = mm.CustomExternalForce('k*periodicdistance(x, y, z, x0, y0, z0)^2')
    else:
        # For non-periodic systems...
        RestraintForce = mm.CustomExternalForce('k*((x-x0)^2+(y-y0)^2+(z-z0)^2)')

    SpringConstant = SpringConstantInKjoules * mm.unit.kilojoules_per_mole/mm.unit.nanometer**2
    RestraintForce.addGlobalParameter('k', SpringConstant)

    RestraintForce.addPerParticleParameter('x0')
    RestraintForce.addPerParticleParameter('y0')
    RestraintForce.addPerParticleParameter('z0')

    for Atom in AtomList:
        RestraintForce.addParticle(Atom.index, Positions[Atom.index])

def GetAtoms(Topology, CAlphaProteinStatus, ResidueNames, Negate):
    """Get a list of atoms in the specified residue names. You may set
    CAlphaProteinStatus flag to True to only retrieve CAlpha atoms from
    the residues. In addition, you may negate residue name match using
    Negate flag.
    
    Arguments:
        Topology (object): OpenMM topology object.
        CAlphaProteinStatus (bool): Get CAlpha atoms only.
        ResidueNames (list): List of residue names.
        Negate (bool): Negate residue name match.

    Returns:
        None or List of OpenMM atom objects.

    """
    AtomList = []
    for Chain in Topology.chains():
        for Residue in Chain.residues():
            if CAlphaProteinStatus:
                if Residue.name in ResidueNames:
                    for Atom in Residue.atoms():
                        if _MatchName(Atom.name, ["CA"], Negate):
                            AtomList.append(Atom)
            else:
                if _MatchName(Residue.name, ResidueNames, Negate):
                    AtomList.extend(Residue.atoms())

    if len(AtomList) == 0:
        AtomList = None

    return AtomList

def _MatchName(Name, NamesList, Negate = False):
    """Match name to the names in a list."""

    Status = True if Name in NamesList else False
    if Negate:
        Status = not Status

    return Status

def DoesSystemUsesPeriodicBoundaryConditions(System):
    """Check for the use of periodic boundary conditions in a system.
    
    Arguments:
        System (object): OpenMM system object.

    Returns:
        bool : True - Uses periodic boundary conditions; Otherwise, false. 

    """

    try:
        Status = True if System.usesPeriodicBoundaryConditions() else False
    except Exception:
        Status = False

    return Status

def DoesAtomListOverlapWithSystemConstraints(System, AtomList):
    """Check for the overlap of specified atoms with the atoms involved
    in system constraints.
    
    Arguments:
        System (object): OpenMM system object.
        AtomList (list): List of OpenMM atom objects.

    Returns:
        bool : True - Overlap with system constraints; Otherwise, false. 

    """

    NumConstraints = System.getNumConstraints()
    if NumConstraints == 0:
        return False

    if AtomList is None:
        return False

    AtomListIndices = [Atom.index for Atom in AtomList]

    for Index in range(NumConstraints):
        Particle1Index, Particle2Index, Distance = System.getConstraintParameters(Index)
        if Particle1Index in AtomListIndices or Particle2Index in AtomListIndices:
            return True

    return False

def DoAtomListsOverlap(AtomList1, AtomList2):
    """Check for the overlap of atoms in the specified atom lists.
    
    Arguments:
        AtomList1 (list): List of OpenMM atom objects.
        AtomList2 (list): List of OpenMM atom objects.

    Returns:
        bool : True - Overlap between atoms lists; Otherwise, false. 

    """

    if AtomList1 is None or AtomList2 is None:
        return False

    AtomList1Indices = [Atom.index for Atom in AtomList1]

    for Atom in AtomList2:
        if Atom.index in AtomList1Indices:
            return True

    return False

def DoesSystemContainWater(Topology, WaterResidueNames = ['HOH']):
    """Check for the presence of water residues in a system.
    
    Arguments:
        Topology (object): OpenMM modeller topology object.
        WaterResidueNames (list): List of water residue names.

    Returns:
        bool : True - Contains water; Otherwise, false. 

    """

    Status = False
    for Residue in Topology.residues():
        if Residue.name in WaterResidueNames:
            Status = True
            break

    return Status

def ReadPDBFile(PDBFile):
    """Read molecule from a PDB file.
    
    The supported PDB file formats are pdb and cif.

    Arguments:
        PDBFile (str): Name of PDB file.

    Returns:
        object: OpenMM PDBFile or PDBFilex object.

    """

    FileDir, FileName, FileExt = MiscUtil.ParseFileName(PDBFile)
    if re.match("^pdb$", FileExt, re.I):
        PDBHandle = mm.app.PDBFile(PDBFile)
    elif re.match("^cif$", FileExt, re.I):
        PDBHandle = mm.app.PDBxFile(PDBFile)
    else:
        MiscUti.PrintError("Failed to read PDB file. Invalid PDB file format %s...\n" % PDBFile)

    return PDBHandle

def WritePDBFile(PDBFile, Topology, Positions, KeepIDs = True):
    """Write a PDB file.
    
    The supported PDB file formats are pdb and cif.
    
    Arguments:
        PDBFile (str): Name of PDB file.
        Topology (object): Topology OpenMM object.
        Positions (object): Positions OpenMM object.
        KeepIDs (bool): Keep existing residue and chain IDs.

    Returns:
        None

    """

    FileDir, FileName, FileExt = MiscUtil.ParseFileName(PDBFile)
    if re.match("^pdb$", FileExt, re.I):
        mm.app.PDBFile.writeFile(Topology, Positions, PDBFile, KeepIDs)
    elif re.match("^cif$", FileExt, re.I):
        mm.app.PDBxFile.writeFile(Topology, Positions, PDBFile, KeepIDs)
    else:
        MiscUti.PrintError("Failed to write PDB file. Invalid PDB file format %s...\n" % PDBFile)

def WriteSimulationStatePDBFile(Simulation, PDBFile, KeepIDs = True):
    """Write a PDB file for current simulation state.
    
    The supported PDB file formats are pdb and cif.
    
    Arguments:
        Simulation (object): OpenMM simulation object.
        PDBFile (str): Name of PDB fil.
        KeepIDs (bool): Keep existing residue and chain IDs.

    Returns:
        None

    """

    CurrentPositions = Simulation.context.getState(getPositions = True).getPositions()
    WritePDBFile(PDBFile, Simulation.topology, CurrentPositions, KeepIDs)

def ReadSmallMoleculeFile(FileName):
    """Read small molecule file using OpenFF toolkit.
    
    Arguments:
        FileName (str): Small molecule file name.

    Returns:
        None or OpenFF tookit molecule object.

    """

    try:
        SmallMol = ff.toolkit.Molecule.from_file(FileName)
    except Exception as ErrMsg:
        SmallMol = None
        MiscUtil.PrintInfo("")
        MiscUtil.PrintWarning("OpenFF.toolkit.Molecule.from_file() failed: %s" % ErrMsg)
        MiscUtil.PrintInfo("")

    return SmallMol

def MergeSmallMoleculeWithMacromolecule(ModellerHandle, SmallMol, SmallMolID = "LIG"):
    """Merge small molecule with macromolecule data contained in a modeller object and
    assign a three letter small molecule residue name to the merged small molecule. 

    Arguments:
        ModellerHandle (object): OpenMM modeller object.
        SmallMol (object): OpenFF tookit molecule object.
        SmallMolID (str): Three letter residue name for small molecule.

    Returns:
        None

    """

    SmallMolToplogy = SmallMol.to_topology()
    SmallMolOpenMMTopology = SmallMolToplogy.to_openmm()
    SmallMolOpenMMPositions = SmallMolToplogy.get_positions().to_openmm()

    # Set small molecule residue name to LIG...
    for Chain in SmallMolOpenMMTopology.chains():
        for Residue in Chain.residues():
            Residue.name = SmallMolID

    ModellerHandle.add(SmallMolOpenMMTopology, SmallMolOpenMMPositions)

def GetFormattedTotalSimulationTime(StepSize, Steps):
    """Get formatted total simulation time with appropriate time units.
    parameter names and values.
    
    Arguments:
        StepSize (object): OpenMM quantity object.
        Steps (int): Number of steps.

    Returns:
        str: Total time.

    """

    TotalTime = StepSize * Steps
    TotalTimeValue = TotalTime.value_in_unit(mm.unit.femtoseconds)

    if TotalTimeValue < 1e3:
        TotalTimeUnits = "fs"
        TotalTime = TotalTime.value_in_unit(mm.unit.femtoseconds)
    elif TotalTimeValue < 1e6:
        TotalTimeUnits = "ps"
        TotalTime = TotalTime.value_in_unit(mm.unit.picoseconds)
    elif TotalTimeValue < 1e9:
        TotalTimeUnits = "ns"
        TotalTime = TotalTime.value_in_unit(mm.unit.nanoseconds)
    elif TotalTimeValue < 1e12:
        TotalTimeUnits = "us"
        TotalTime = TotalTime.value_in_unit(mm.unit.microseconds)
    else:
        TotalTimeUnits = "ms"
        TotalTime = TotalTime.value_in_unit(mm.unit.milliseconds)

    TotalTime = "%.2f %s" % (TotalTime, TotalTimeUnits)

    return TotalTime

def ProcessOptionOpenMMRestartParameters(ParamsOptionName, ParamsOptionValue, OutfilePrefix, ParamsDefaultInfo = None):
    """Process parameters for restart option and return a map containing processed
    parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    to setup platform.
    
    The supported parameter names along with their default and possible
    values are shown below:
    
        finalStateFile, <OutfilePrefix>_FinalState.<chk>  [ Possible values:
            Valid final state checkpoint or XML filename ]
        dataAppend, yes [ Possible values: yes or no]
            
    A brief description of parameters is provided below:
    
    finalStateFile: Final state checkpoint or XML file
    
    dataAppend: Append data to existing trajectory and data log files during the
    restart of a simulation using a previously saved  final state checkpoint or
    XML file.

    Arguments:
        ParamsOptionName (str): Command line OpenMM restart option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        OutfilePrefix (str): Prefix for output files.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"FinalStateFile": "auto", "DataAppend": True}

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMRestartParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        if re.match("^FinalStateFile$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not os.path.exists(Value):
                    MiscUtil.PrintError("The file name specified, %s, for parameter name, %s, using option \"%s\" doesn't exist.\n." % (Value, Name, ParamsOptionName))
                if not MiscUtil.CheckFileExt(Value, "chk xml"):
                    MiscUtil.PrintError("The file name specified, %s, for parameter name, %s, using option \"%s\" is not valid file. Supported file formats: chk or xml\n." % (Value, Name, ParamsOptionName))
                ParamValue = Value
        elif re.match("^DataAppend$", ParamName, re.I):
            if not re.match("^(yes|no|true|false)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: yes or no" % (Value, Name, ParamsOptionName))
            ParamValue = True if re.match("^(yes|true)$", Value, re.I) else False
        else:
            ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMRestartParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix)

    return ParamsInfo

def _ProcessOptionOpenMMRestartParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix):
    """Process parameters with possible auto values and perform validation.
    """

    FinalStateFileCheckpointMode = False
    FinalStateFileXMLMode = False

    ParamName = "FinalStateFile"
    FinalStateFile = ParamsInfo[ParamName]
    if re.match("^auto$", FinalStateFile, re.I):
        FinalStateFile = "%s_FinalState.chk" % OutfilePrefix
        FinalStateFileCheckpointMode = True
    else:
        if MiscUtil.CheckFileExt(FinalStateFile, "chk"):
            FinalStateFileCheckpointMode = True
        elif MiscUtil.CheckFileExt(FinalStateFile, "xml"):
            FinalStateFileXMLMode = True
        else:
            MiscUtil.PrintError("The file name specified, %s, for parameter name, %s, using option \"%s\ is not valid. Supported file formats: chk or xml\n." % (FinalStateFile, ParamName, ParamsOptionName))

    ParamsInfo["FinalStateFile"] = FinalStateFile
    ParamsInfo["FinalStateFileCheckpointMode"] = FinalStateFileCheckpointMode
    ParamsInfo["FinalStateFileXMLMode"] = FinalStateFileXMLMode

def ProcessOptionOpenMMSystemParameters(ParamsOptionName, ParamsOptionValue, ParamsDefaultInfo = None):
    """Process parameters for system option and return a map containing processed
    parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    to setup platform.
    
    The supported parameter names along with their default and possible
    values are shown below:
    
        constraints, BondsInvolvingHydrogens [ Possible values: None,
            WaterOnly, BondsInvolvingHydrogens, AllBonds, or
            AnglesInvolvingHydrogens ]
        constraintErrorTolerance, 0.000001
        ewaldErrorTolerance, 0.0005
        
        nonbondedMethodPeriodic, PME [ Possible values: NoCutoff,
            CutoffNonPeriodic, or PME ]
        nonbondedMethodNonPeriodic, NoCutoff [ Possible values:
            NoCutoff or CutoffNonPeriodic]
        nonbondedCutoff, 1.0 [ Units: nm ]
        
        hydrogenMassRepartioning, yes [ Possible values: yes or no ]
        hydrogenMass, 1.5 [ Units: amu]
        
        removeCMMotion, yes [ Possible values: yes or no ]
        rigidWater, auto [ Possible values: yes or no. Default: 'No' for
            'None' value of constraints; Otherwise, yes ]
        
    A brief description of parameters is provided below:
    
    constraints: Type of system constraints to use for simulation. These constraints
    are different from freezing and restraining of any atoms in the system.
    
    constraintErrorTolerance: Distance tolerance for constraints as a fraction
    of the constrained distance.
    
    ewaldErrorTolerance: Ewald error tolerance for a periodic system.
    
    nonbondedMethodPeriodic: Nonbonded method to use during the calculation of
    long range interactions for a periodic system.
    
    nonbondedMethodNonPeriodic: Nonbonded method to use during the calculation
    of long range interactions for a non-periodic system.
    
    nonbondedCutoff: Cutoff distance to use for long range interactions in both
    perioidic non-periodic systems.
    
    hydrogenMassRepartioning: Use hydrogen mass repartioning. It increases the
    mass of the hydrogen atoms attached to the heavy atoms and decreasing the
    mass of the bonded heavy atom to maintain constant system mass. This allows
    the use of larger integration step size (4 fs) during a simulation.
    
    hydrogenMass: Hydrogen mass to use during repartioning.
    
    removeCMMotion: Remove all center of mass motion at every time step.
    
    rigidWater: Keep water rigid during a simulation. This is determined
    automatically based on the value of 'constraints' parameter.

    Arguments:
        ParamsOptionName (str): Command line OpenMM system option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"Constraints": "BondsInvolvingHydrogens", "ConstraintErrorTolerance": 0.000001, "EwaldErrorTolerance": 0.0005, "NonbondedMethodPeriodic": "PME", "NonbondedMethodNonPeriodic": "NoCutoff", "NonbondedCutoff": 1.0,  "HydrogenMassRepartioning": True, "HydrogenMass": 1.5, "RemoveCMMotion":  True, "RigidWater": "auto"}

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMSystemParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        if re.match("^Constraints$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not re.match("^(None|WaterOnly|BondsInvolvingHydrogens|AllBonds|AnglesInvolvingHydrogens)$", Value, re.I):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: None, WaterOnly, BondsInvolvingHydrogens, AllBonds, or AnglesInvolvingHydrogens." % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^(ConstraintErrorTolerance|EwaldErrorTolerance|NonbondedCutoff|HydrogenMass)$", ParamName, re.I):
            if not MiscUtil.IsFloat(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
            Value = float(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^NonbondedMethodPeriodic$", ParamName, re.I):
            if not re.match("^(NoCutoff|CutoffPeriodic|PME)$", Value, re.I):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: NoCutoff, CutoffPeriodic, or PME" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^NonbondedMethodNonPeriodic$", ParamName, re.I):
            if not re.match("^(NoCutoff|CutoffNonPeriodic)$", Value, re.I):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: NoCutoff or CutoffNonPeriodic" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^(HydrogenMassRepartioning|removeCMMotion)$", ParamName, re.I):
            if not re.match("^(yes|no|true|false)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: yes or no" % (Value, Name, ParamsOptionName))
            ParamValue = True if re.match("^(yes|true)$", Value, re.I) else False
        elif re.match("^RigidWater$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not re.match("^(yes|no|true|false)$", Value, re.I):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: yes or no" % (Value, Name, ParamsOptionName))
                ParamValue = True if re.match("^(yes|true)$", Value, re.I) else False
        else:
            ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMSystemParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)

    return ParamsInfo

def _ProcessOptionOpenMMSystemParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue):
    """Process parameters with possible auto values and perform validation.
    """

    for ParamName in ["NonbondedMethodPeriodic", "NonbondedMethodNonPeriodic"]:
        ParamValue = ParamsInfo[ParamName]
        if re.match("^NoCutoff$", ParamValue, re.I):
            ParamValue = mm.app.NoCutoff
        elif re.match("^CutoffNonPeriodic$", ParamValue, re.I):
            ParamValue = mm.app.CutoffNonPeriodic
        elif re.match("^CutoffPeriodic$", ParamValue, re.I):
            ParamValue = mm.app.CutoffPeriodic
        elif re.match("^PME$", ParamValue, re.I):
            ParamValue = mm.app.PME
        else:
            MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: NoCutoff, CutoffNonPeriodic, CutoffPeriodic, or PME" % (ParamValue, ParamName, ParamsOptionName))
        ParamsInfo[ParamName] = ParamValue

    ParamName = "Constraints"
    ParamValue = ParamsInfo[ParamName]
    ConstraintsValue = None
    RigidWaterValue = False
    if re.match("^None$", ParamValue, re.I):
        ConstraintsValue = None
        RigidWaterValue = False
    elif re.match("^WaterOnly$", ParamValue, re.I):
        ConstraintsValue = None
        RigidWaterValue = True
    elif re.match("^BondsInvolvingHydrogens$", ParamValue, re.I):
        ConstraintsValue = mm.app.HBonds
        RigidWaterValue = True
    elif re.match("^AllBonds$", ParamValue, re.I):
        ConstraintsValue = mm.app.AllBonds
        RigidWaterValue = True
    elif re.match("^AnglesInvolvingHydrogens$", ParamValue, re.I):
        ConstraintsValue = mm.app.HAngles
        RigidWaterValue = True
    else:
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: None, WaterOnly, BondsInvolvingHydrogens, AllBonds, or AnglesInvolvingHydrogens." % (ParamValue, ParamName, ParamsOptionName))

    ParamsInfo[ParamName] = ConstraintsValue

    ParamName = "RigidWater"
    ParamValue = "%s" % ParamsInfo[ParamName]
    if re.match("^auto$", ParamValue, re.I):
        ParamsInfo[ParamName] = RigidWaterValue

def ProcessOptionOpenMMIntegratorParameters(ParamsOptionName, ParamsOptionValue, ParamsDefaultInfo = None, HydrogenMassRepartioningStatus = False):
    """Process parameters for integrator option and return a map containing processed
    parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    to setup platform.
    
    The supported parameter names along with their default and possible
    values are shown below:
    
        integrator, LangevinMiddle [ Possible values: LangevinMiddle,
            Langevin, NoseHoover, Brownian ]
        
        randomSeed, auto [ Possible values: > 0 ]
        
        frictionCoefficient, 1.0 [ Units: 1/ps ]
        stepSize, auto [ Units: fs; Default value: 4 fs during yes value of
            hydrogen mass repartioning with no freezing/restraining of atoms;
            otherwsie, 2 fs ] 
        temperature, 300.0 [ Units: kelvin ]
        
        barostat, MonteCarlo [ Possible values: MonteCarlo or
            MonteCarloMembrane ]
        barostatInterval, 25
        pressure, 1.0 [ Units: atm ]
        
        Parameters used only for MonteCarloMembraneBarostat with default
        values corresponding to Amber forcefields:
        
        surfaceTension, 0.0 [ Units: atm*A. It is automatically converted 
            into OpenMM default units of atm*nm before its usage.  ]
        xymode,  Isotropic [ Possible values: Anisotropic or  Isotropic ]
        zmode,  Free [ Possible values: Free or  Fixed ]
    
    A brief description of parameters is provided below:
        
    integrator: Type of integrator
    
    randomSeed: Random number seed for barostat and integrator. Not supported
        NoseHoover integrator.
    
    frictionCoefficient: Friction coefficient for coupling the system to the heat
    bath.
    
    stepSize: Simulation time step size.
    
    temperature: Simulation temperature.
    
    barostat: Barostat type.
    
    barostatInterval: Barostat interval step size during NPT simulation for
    applying Monte Carlo pressure changes.
    
    pressure: Pressure during NPT simulation. 
    
    surfaceTension: Surface tension acting on the system.
    
    xymode: Behavior along X and Y axes. You may allow the X and Y axes
    to vary independently of each other or always scale them by the same
    amount to keep the ratio of their lengths constant.
    
    zmode: Beahvior along Z axis. You may allow the Z axis to vary
    independently of the other axes or keep it fixed.

    Arguments:
        ParamsOptionName (str): Command line OpenMM integrator option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"Integrator": "LangevinMiddle", "RandomSeed":  "auto", "FrictionCoefficient": 1.0, "StepSize": "auto", "Temperature": 300.0, "Barostat": "MonteCarlo", "BarostatInterval": 25, "Pressure": 1.0, "SurfaceTension": 0.0, "XYMode": "Isotropic", "ZMode": "Free"}

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMIntegratorParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, HydrogenMassRepartioningStatus)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        if re.match("^Integrator$", ParamName, re.I):
            if not re.match("^(LangevinMiddle|Langevin|NoseHoover|Brownian)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: LangevinMiddle,Langevin, NoseHoover, or Brownian." % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^(FrictionCoefficient|Pressure)$", ParamName, re.I):
            if not MiscUtil.IsFloat(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
            Value = float(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^StepSize$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.IsFloat(Value):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
                Value = float(Value)
                if Value <= 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
                ParamValue = Value
        elif re.match("^(Temperature|SurfaceTension)$", ParamName, re.I):
            if not MiscUtil.IsFloat(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
            Value = float(Value)
            if Value < 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: >= 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^BarostatInterval$", ParamName, re.I):
            if not MiscUtil.IsInteger(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be an integer.\n" % (Value, ParamName, ParamsOptionName))
            Value = int(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^Barostat$", ParamName, re.I):
            if not re.match("^(MonteCarlo|MonteCarloMembrane)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: MonteCarlo or MonteCarloMembrane" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^XYMode$", ParamName, re.I):
            if not re.match("^(Anisotropic|Isotropic)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: Anisotropic or  Isotropic" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^ZMode$", ParamName, re.I):
            if not re.match("^(Free|Fixed)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: Fixed or Free" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^RandomSeed$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.IsInteger(Value):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be an integer.\n" % (Value, ParamName, ParamsOptionName))
                Value = int(Value)
                if Value <= 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
                ParamValue = Value
        else:
            ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMIntegratorParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, HydrogenMassRepartioningStatus)

    return ParamsInfo

def _ProcessOptionOpenMMIntegratorParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, HydrogenMassRepartioningStatus):
    """Process parameters with possible auto values and perform validation.
    """

    ParamName = "StepSize"
    ParamValue = "%s" % ParamsInfo[ParamName]
    ParamsInfo["StepSizeSpecified"] = ParamValue
    if re.match("^auto$", ParamValue, re.I):
        ParamValue = 4.0 if HydrogenMassRepartioningStatus else 2.0
        ParamsInfo[ParamName] = ParamValue

    ParamName = "RandomSeed"
    ParamValue = "%s" % ParamsInfo[ParamName]
    ParamsInfo["RandomSeedSpecified"] = ParamValue
    if re.match("^auto$", ParamValue, re.I):
        ParamsInfo[ParamName] = None

def ProcessOptionOpenMMSimulationParameters(ParamsOptionName, ParamsOptionValue, ParamsDefaultInfo = None):
    """Process parameters for simulation option and return a map containing processed
    parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    to setup platform.
    
    The supported parameter names along with their default and possible
    values are shown below:
    
        steps, 1000000 [ Possible values: > 0 ] 
        
        minimization, yes [ Possible values: yes or no ] 
        minimizationMaxSteps, auto  [ Possible values: >= 0. The value of
            zero implies until the minimization is converged. ]
        minimizationTolerance, 0.25  [ Units: kcal/mol/A. The default value
            0.25, corresponds to OpenMM default of value of 10.04
            kjoules/mol/nm. It is automatically converted into OpenMM
            default units before its usage. ]
        
        equilibration, yes [ Possible values: yes or no ] 
        equilibrationSteps, 1000  [ Possible values: > 0 ]
    
    A brief description of parameters is provided below:
    
    steps: Number of steps for production run.
    
    equilibration: Perform equilibration before the production run.
    
    equilibrationSteps: Number of steps for equilibration.
    
    minimizationMaxSteps: Maximum number of minimization steps. The value
    of zero implies until the minimization is converged.
    
    minimizationTolerance: Energy convergence tolerance during minimization.

    minimization: Perform minimization before equilibration and production run.

    Arguments:
        ParamsOptionName (str): Command line OpenMM simulation option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"Steps": 1000000, "Minimization": True, "MinimizationMaxSteps": "auto", "MinimizationTolerance": 0.24, "Equilibration": True, "EquilibrationSteps": 1000}

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMSimulationParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        if re.match("^(Steps|EquilibrationSteps)$", ParamName, re.I):
            if not MiscUtil.IsInteger(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be an integer.\n" % (Value, ParamName, ParamsOptionName))
            Value = int(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^(Minimization|Equilibration)$", ParamName, re.I):
            if not re.match("^(yes|no|true|false)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: yes or no" % (Value, Name, ParamsOptionName))
            ParamValue = True if re.match("^(yes|true)$", Value, re.I) else False
        elif re.match("^MinimizationMaxSteps$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.IsInteger(Value):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be an integer.\n" % (Value, ParamName, ParamsOptionName))
                Value = int(Value)
                if Value < 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: >= 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^MinimizationTolerance$", ParamName, re.I):
            if not MiscUtil.IsFloat(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
            Value = float(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        else:
            ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMSimulationParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)

    return ParamsInfo

def _ProcessOptionOpenMMSimulationParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue):
    """Process parameters with possible auto values and perform validation.
    """

    ParamName = "MinimizationMaxSteps"
    ParamValue = "%s" % ParamsInfo[ParamName]
    if re.match("^auto$", ParamValue, re.I):
        ParamsInfo[ParamName] = 0

def ProcessOptionOpenMMOutputParameters(ParamsOptionName, ParamsOptionValue, OutfilePrefix, ParamsDefaultInfo = None):
    """Process parameters for output option and return a map containing processed
    parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    to setup platform.
    
    The supported parameter names along with their default and possible
    values are shown below:
    
        checkpoint, no  [ Possible values: yes or no ]
        checkpointFile, auto  [ Default: <OutfilePrefix>.chk ]
        checkpointSteps, 10000
        
        dataOutType, auto [ Possible values: A space delimited list of valid
            parameter names.
            NPT simulation default: Density Step Speed Progress
                  PotentialEnergy Temperature Time.
            NVT simulation default: Step Speed Progress PotentialEnergy
                Temperature Time Volumne
            Other valid names: ElapsedTime RemainingTime KineticEnergy
                TotalEnergy  ]
        
        dataLog, yes  [ Possible values: yes or no ]
        dataLogFile, auto  [ Default: <OutfilePrefix>.csv ]
        dataLogSteps, 1000
        
        dataStdout, no  [ Possible values: yes or no ]
        dataStdoutSteps, 1000
        
        minimizationDataSteps, 100
        minimizationDataStdout, no  [ Possible values: yes or no ]
        minimizationDataLog, no  [ Possible values: yes or no ]
        minimizationDataLogFile, auto  [ Default:
            <OutfilePrefix>_MinimizationOut.csv ]
        minimizationDataOutType, auto [ Possible values: A space delimited
            list of valid parameter names.  Default: SystemEnergy
            RestraintEnergy MaxConstraintError.
            Other valid names: RestraintStrength ]
        
        pdbOutFormat, PDB  [ Possible values: PDB or CIF ]
        pdbOutKeepIDs, yes  [ Possible values: yes or no ]
        
        pdbOutMinimized, no  [ Possible values: yes or no ]
        pdbOutEquilibrated, no  [ Possible values: yes or no ]
        pdbOutFinal, no  [ Possible values: yes or no ]
        
        saveFinalStateCheckpoint, yes  [ Possible values: yes or no ]
        saveFinalStateCheckpointFile, auto  [ Default:
            <OutfilePrefix>_FinalState.chk ]
        saveFinalStateXML, no  [ Possible values: yes or no ]
        saveFinalStateXMLFile, auto  [ Default:
            <OutfilePrefix>_FinalState.xml]
        
        traj, yes  [ Possible values: yes or no ]
        trajFile, auto  [ Default: <OutfilePrefix>.<TrajFormat> ]
        trajFormat, DCD  [ Possible values: DCD or XTC ]
        trajSteps, 10000
        
        xmlSystemOut, no  [ Possible values: yes or no ]
        xmlSystemFile, auto  [ Default: <OutfilePrefix>_System.xml ]
        xmlIntegratorOut, no  [ Possible values: yes or no ]
        xmlIntegratorFile, auto  [ Default: <OutfilePrefix>_Integrator.xml ]
            
    A brief description of parameters is provided below:
        
        checkpoint: Write intermediate checkpoint file.
        checkpointFile: Intermediate checkpoint file name.
        checkpointSteps: Frequency of writing intermediate checkpoint file.
        
        dataOutType: Type of data to write to stdout and log file.
        
        dataLog: Write data to log file.
        dataLogFile: Data log file name.
        dataLogSteps: Frequency of writing data to log file.
        
        dataStdout: Write data to stdout.
        dataStdoutSteps: Frequency of writing data to stdout.
        
        minimizationDataSteps: Frequency of writing data to stdout and log file.
        minimizationDataStdout: Write data to stdout.
        minimizationDataLog: Write data to log file.
        minimizationDataLogFile: Data log fie name.
        minimizationDataOutType: Type of data to write to stdout and log file.
        
        pdbOutFormat: Format of output PDB files.
        pdbOutKeepIDs: Keep existing chain and residue IDs.
        
        pdbOutMinimized: Write PDB file after minimization.
        pdbOutEquilibrated: Write PDB file after equilibration.
        pdbOutFinal: Write final PDB file after production run.
        
        saveFinalStateCheckpoint: Save final state checkpoint file.
        saveFinalStateCheckpointFile: Name of final state checkpoint file.
        saveFinalStateXML: Save final state XML file.
        saveFinalStateXMLFile: Name of final state XML file.
        
        traj: Write out trajectory file.
        trajFile: Trajectory file name.
        trajFormat: Trajectory file format.
        trajSteps: Frequency of writing trajectory file.
        
        xmlSystemOut: Write system XML file.
        xmlSystemFile: System XML file name.
        xmlIntegratorOut: Write integrator XML file.
        xmlIntegratorFile: Integrator XML file name.

    Arguments:
        ParamsOptionName (str): Command line OpenMM system option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"Checkpoint": False, "CheckpointFile": "auto",  "CheckpointSteps": 10000,
                  "DataOutType": "auto",
                  "DataLog": True, "DataLogFile": "auto", "DataLogSteps": 1000,
                  "DataStdout":  False, "DataStdoutSteps": 1000,
                  "MinimizationDataSteps": 100, "MinimizationDataStdout": False, "MinimizationDataLog": False, "MinimizationDataLogFile": "auto", "MinimizationDataOutType": "auto",
                  "PDBOutFormat": "PDB", "PDBOutKeepIDs": True, "PDBOutMinimized": False, "PDBOutEquilibrated": False, "PDBOutFinal": False, 
                  "SaveFinalStateCheckpoint": True, "SaveFinalStateCheckpointFile": "auto", "SaveFinalStateXML": False, "SaveFinalStateXMLFile": "auto",
                  "Traj": True,  "TrajFile": "auto", "TrajFormat": "DCD", "TrajSteps": 10000,
                  "XmlSystemOut": False, "XmlSystemFile": "auto",  "XmlIntegratorOut": False, "XmlIntegratorFile": "auto"}

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMOutputParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        if re.match("^(Checkpoint|DataLog|DataStdout|MinimizationDataStdout|MinimizationDataLog|PDBOutKeepIDs|SaveFinalStateCheckpoint|SaveFinalStateXML|Traj|XmlSystemOut|XmlIntegratorOut|PDBOutMinimized|PDBOutEquilibrated|PDBOutFinal)$", ParamName, re.I):
            if not re.match("^(yes|no|true|false)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: yes or no" % (Value, Name, ParamsOptionName))
            ParamValue = True if re.match("^(yes|true)$", Value, re.I) else False
        elif re.match("^(CheckpointFile|SaveFinalStateCheckpointFile)$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.CheckFileExt(Value, "chk"):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported file format: chk" % (Value, Name, ParamsOptionName))
                ParamValue = Value
        elif re.match("^(DataLogFile|MinimizationDataLogFile)$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.CheckFileExt(Value, "csv"):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported file formats: csv" % (Value, Name, ParamsOptionName))
                ParamValue = Value
        elif re.match("^(SaveFinalStateXMLFile|XmlSystemFile|XmlIntegratorFile)$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.CheckFileExt(Value, "xml"):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported file format: xml" % (Value, Name, ParamsOptionName))
                ParamValue = Value
        elif re.match("^TrajFile$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.CheckFileExt(Value, "dcd xtc"):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported file formats: dcd xtc" % (Value, Name, ParamsOptionName))
                ParamValue = Value
        elif re.match("^(CheckpointSteps|DataLogSteps|DataStdoutSteps|MinimizationDataSteps|TrajSteps)$", ParamName, re.I):
            if not MiscUtil.IsInteger(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be an integer.\n" % (Value, ParamName, ParamsOptionName))
            Value = int(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^DataOutType$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                ValueTypes = Value.split()
                if len(ValueTypes) == 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain a space delimited list of valid values.\n" % (Value, ParamName, ParamsOptionName))
                ValueTypesSpecified = []
                for ValueType in ValueTypes:
                    if not re.match("^(Step|Speed|Progress|PotentialEnergy|Temperature|ElapsedTime|RemainingTime|Time|KineticEnergy|TotalEnergy|Volume|Density)$", ValueType, re.I):
                        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: Step, Speed, Progress, PotentialEnergy, Temperature, ElapsedTime, RemainingTime, Time, KineticEnergy, TotalEnergy, Volume, or Density" % (ValueType, Name, ParamsOptionName))
                    if ValueType in ValueTypesSpecified:
                        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It has already been specified." % (ValueType, Name, ParamsOptionName))
                    ValueTypesSpecified.append(ValueType)
                ParamsInfo["DataOutTypeList"] = ValueTypes
        elif re.match("^MinimizationDataOutType$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                ValueTypes = Value.split()
                if len(ValueTypes) == 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain a space delimited list of valid values.\n" % (Value, ParamName, ParamsOptionName))
                ValueTypesSpecified = []
                for ValueType in ValueTypes:
                    if not re.match("^(SystemEnergy|RestraintEnergy|RestraintStrength|MaxConstraintError)$", ValueType, re.I):
                        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: SystemEnergy, RestraintEnergy, RestraintStrength, or MaxConstraintError" % (ValueType, Name, ParamsOptionName))
                    if ValueType in ValueTypesSpecified:
                        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It has already been specified." % (ValueType, Name, ParamsOptionName))
                    ValueTypesSpecified.append(ValueType)
                ParamsInfo["MinimizationDataOutTypeList"] = ValueTypes
        elif re.match("^PDBOutFormat$", ParamName, re.I):
            if not re.match("^(PDB|CIF)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: PDB or CIF" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^TrajFormat$", ParamName, re.I):
            if not re.match("^(DCD|XTC)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: DCD or XTC" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        else:
            ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMOutputParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix)

    return ParamsInfo

def _ProcessOptionOpenMMOutputParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix):
    """Process parameters with possible auto values and perform validation.
    """

    # Use comma as a delimiter...
    ParamsInfo["DataOutDelimiter"] = ","
    ParamsInfo["DataOutfileExt"] = "csv"

    ParamName = "TrajFormat"
    ParamValue = ParamsInfo[ParamName]
    if re.match("^DCD$", ParamValue, re.I):
        TrajFileExt = "dcd"
    elif re.match("^XTC$", ParamValue, re.I):
        TrajFileExt = "xtc"
    else:
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: DCD or XTC" % (ParamValue, ParamName, ParamsOptionName))
    ParamsInfo["TrajFileExt"] = TrajFileExt

    ParamName = "PDBOutFormat"
    ParamValue = ParamsInfo[ParamName]
    if re.match("^PDB$", ParamValue, re.I):
        PDBOutfileExt = "pdb"
    elif re.match("^CIF$", ParamValue, re.I):
        PDBOutfileExt = "cif"
    else:
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: PDB or CIF" % (ParamValue, ParamName, ParamsOptionName))
    ParamsInfo["PDBOutfileExt"] = PDBOutfileExt

    _ProcessFileNamesOutputPatramaters(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix)
    _ProcessDataOutTypeOutputParameter(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix)
    _ProcessMinimizationDataOutTypeOutputParameter(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix)
    
def _ProcessFileNamesOutputPatramaters(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix):
    """Process output parameters corresponding to file names."""

    OutfileExt = ParamsInfo["DataOutfileExt"]
    TrajFileExt = ParamsInfo["TrajFileExt"]
    OutfilesSuffixAndExtMap = {"CheckpointFile": ["", "chk"], "DataLogFile": ["", OutfileExt], "MinimizationDataLogFile": ["Minimization", OutfileExt], "SaveFinalStateCheckpointFile": ["FinalState", "chk"], "SaveFinalStateXMLFile": ["FinalState", "xml"], "TrajFile": ["", TrajFileExt], "XmlSystemFile": ["System", "xml"], "XmlIntegratorFile": ["Integrator", "xml"]}
    OutfileNames = []
    for ParamName in OutfilesSuffixAndExtMap:
        ParamValue = ParamsInfo[ParamName]
        if re.match("^auto$", ParamValue, re.I):
            DataOutfileSuffix, DataOutfileExt = OutfilesSuffixAndExtMap[ParamName]
            if len(DataOutfileSuffix):
                DataOutfileSuffix = "_%s" % DataOutfileSuffix
            ParamValue = "%s%s.%s" % (OutfilePrefix, DataOutfileSuffix, DataOutfileExt)
            ParamsInfo[ParamName] = ParamValue
        else:
            # Check for duplicate output file names...
            if ParamValue not in OutfileNames:
                OutfileNames.append(ParamValue)
            else:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It's a duplicate file name." % (ParamValue, ParamName, ParamsOptionName))

            # Validate specified traj file extension...
            if re.match("^TrajFile$", ParamName, re.I):
                TrajFormat = ParamsInfo["TrajFormat"]
                if not MiscUtil.CheckFileExt(ParamValue, TrajFileExt):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. The file extension must match must match the extenstion, %s, corresponding to trajectory format, %s, speecified using \"--trajFormat\" option." % (ParamValue, ParamName, ParamsOptionName, TrajFileExt, TrajFormat))

def _ProcessDataOutTypeOutputParameter(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix):
    """Process output parameter corresponding to data out type."""

    # Setup data out types...
    DataOutTypeStatusMap = {"Step": False, "Speed": False,  "Progress": False,  "PotentialEnergy": False,  "Temperature": False, "ElapsedTime": False, "RemainingTime": False, "Time": False, "KineticEnergy": False, "TotalEnergy": False,  "Volume": False,  "Density": False}
    CanonicalDataOutTypeMap = {}
    ValidDataOutTypes = []
    for DataOutType in DataOutTypeStatusMap:
        ValidDataOutTypes.append(DataOutType)
        CanonicalDataOutTypeMap[DataOutType.lower()] = DataOutType

    # Process data out types...
    ParamName = "DataOutType"
    ParamValue = ParamsInfo[ParamName]
    DataOutTypeList = []
    if re.match("^auto$", ParamValue, re.I):
        DataOutTypeList = ["Step", "Speed", "Progress", "PotentialEnergy", "Temperature"]
    else:
        if "DataOutTypeList" in ParamsInfo:
            DataOutTypeList = ParamsInfo["DataOutTypeList"]
        else:
            DataOutTypeList = ParamsInfo["DataOutType"].split()
            ParamsInfo["DataOutTypeList"] = DataOutTypeList

    for DataOutType in DataOutTypeList:
        CanonicalDataOutType = DataOutType.lower()
        if CanonicalDataOutType not in CanonicalDataOutTypeMap:
            MiscUtil.PrintError("The parameter value, %s specified for paramaer name, %s, specified using \"%s\" is not a valid name. Supported parameter names: %s" % (DataOutType, ParamName, ParamsOptionName, " ".join(ValidDataOutTypes)))
        
        DataOutType = CanonicalDataOutTypeMap[CanonicalDataOutType]
        DataOutTypeStatusMap[DataOutType] = True

    ParamsInfo["DataOutTypeList"] = DataOutTypeList
    ParamsInfo["DataOutTypeStatusMap"] = DataOutTypeStatusMap

def _ProcessMinimizationDataOutTypeOutputParameter(ParamsInfo, ParamsOptionName, ParamsOptionValue, OutfilePrefix):
    """Process output parameter corresponding to minimization data out type."""

    # Setup mimimization data out types...
    DataOutTypeOpenMMNameMap = {"SystemEnergy": "system energy", "RestraintEnergy": "restraint energy", "RestraintStrength": "restraint strength", "MaxConstraintError": "max constraint error"}

    CanonicalDataOutTypeMap = {}
    ValidDataOutTypes = []
    for DataOutType in DataOutTypeOpenMMNameMap:
        ValidDataOutTypes.append(DataOutType)
        CanonicalDataOutTypeMap[DataOutType.lower()] = DataOutType

    # Process minimization data out types...
    ParamName = "MinimizationDataOutType"
    ParamValue = ParamsInfo[ParamName]
    DataOutTypeList = []
    if re.match("^auto$", ParamValue, re.I):
        DataOutTypeList = ["SystemEnergy", "RestraintEnergy", "MaxConstraintError"]
    else:
        if "MinimizationDataOutTypeList" in ParamsInfo:
            DataOutTypeList = ParamsInfo["MinimizationDataOutTypeList"]
        else:
            DataOutTypeList = ParamsInfo["MinimizationDataOutType"].split()
            ParamsInfo["MinimizationDataOutTypeList"] = DataOutTypeList

    # Set up a list containing OpenMM names for minimization reporter...
    DataOutTypeOpenMMNameList = []
    for DataOutType in DataOutTypeList:
        CanonicalDataOutType = DataOutType.lower()
        if CanonicalDataOutType not in CanonicalDataOutTypeMap:
            MiscUtil.PrintError("The parameter value, %s specified for paramaer name, %s, specified using \"%s\" is not a valid name. Supported parameter names: %s" % (DataOutType, ParamName, ParamsOptionName, " ".join(ValidDataOutTypes)))
        
        DataOutType = CanonicalDataOutTypeMap[CanonicalDataOutType]
        
        DataOutTypeOpenMMName = DataOutTypeOpenMMNameMap[DataOutType]
        DataOutTypeOpenMMNameList.append(DataOutTypeOpenMMName)

    ParamsInfo["MinimizationDataOutTypeList"] = DataOutTypeList
    ParamsInfo["MinimizationDataOutTypeOpenMMNameList"] = DataOutTypeOpenMMNameList
    
def ProcessOptionOpenMMAtomsSelectionParameters(ParamsOptionName, ParamsOptionValue, ParamsDefaultInfo = None):
    """Process parameters for selecting atoms and return a map containing
    processed parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    to select atoms.
    
    The supported parameter names along with their default and possible
    values are shown below:
    
        selection, none [ Possible values: CAlphaProtein, Ions, Ligand,
            Protein, Residues, or Water ]
        selectionSpec, auto [ Possible values: A space delimited list of
            residue names ]
        negate, no [ Possible values: yes or no ]
    
    A brief description of parameters is provided below:
    
    selection: Atom selection to freeze.
    
    selectionSpec: A space delimited list of residue names for selecting atoms.
    You must specify its value during 'Ligand' and 'Protein' value for 'selection'.
    The default values are automatically set for 'CAlphaProtein', 'Ions', 'Protein',
    and 'Water' values of 'selection' as shown below:
    
        CAlphaProtein: List of stadard protein residues from pdbfixer
            for selecting CAlpha atoms.
        Ions: Li Na K Rb Cs Cl Br F I
        Water: HOH
        Protein: List of standard protein residues from pdbfixer.
    
    negate: Negate atom selection match to select atoms for freezing.
    
    In addition, you may specify an explicit space delimited list of residue
    names using 'selectionSpec' for any 'selection". The specified residue
    names are appended to the appropriate default values during the
    selection of atoms for freezing.

    Arguments:
        ParamsOptionName (str): Command line OpenMM selection option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"Selection": None, "SelectionSpec": "auto", "Negate": False}

    if ParamsOptionValue is None:
        return ParamsInfo

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMAtomsSelectionParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue
        if re.match("^Selection$", ParamName, re.I):
            if not re.match("^(CAlphaProtein|Ions|Ligand|Protein|Residues|Water)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: CAlphaProtein, Ions, Ligand, Protein, Residues, or Water" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^SelectionSpec$", ParamName, re.I):
            if not re.match("^(auto|none)$", Value, re.I):
                Values = Value.split()
                if len(Values) == 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain space delimited list of residue names.\n" % (Value, ParamName, ParamsOptionName))
                # Set residues list...
                ParamsInfo["SelectionSpecList"] = Values
        elif re.match("^Negate$", ParamName, re.I):
            if not re.match("^(yes|no|true|false)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: yes or no" % (Value, Name, ParamsOptionName))
            ParamValue = True if re.match("^(yes|true)$", Value, re.I) else False
        else:
            ParamValue = Value
        
        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMAtomsSelectionParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)

    return ParamsInfo

def _ProcessOptionOpenMMAtomsSelectionParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue):
    """Process parameters with possible auto values and perform validation.
    """

    SelectionParamName = "Selection"
    SelectionParamValue = ParamsInfo[SelectionParamName]

    SelectionSpecParamName = "SelectionSpec"
    SelectionSpecParamValue = ParamsInfo[SelectionSpecParamName]

    SelectionSpecList = None if re.match("^(auto|none)$", SelectionSpecParamValue, re.I) else ParamsInfo["SelectionSpecList"]

    ResidueNames = None
    if re.match("^(CAlphaProtein|Protein)$", SelectionParamValue, re.I):
        ResidueNames = pdbfixer.pdbfixer.proteinResidues
        if SelectionSpecList is not None:
            ResidueNames.extend(SelectionSpecList)
    elif re.match("^Ions$", SelectionParamValue, re.I):
        ResidueNames = ["Li", "Na", "K", "Rb", "Cs",  "Cl", "Br", "F", "I"]
        if SelectionSpecList is not None:
            ResidueNames.extend(SelectionSpecList)
    elif re.match("^Ligand$", SelectionParamValue, re.I):
        if SelectionSpecList is None:
            MiscUtil.PrintError("No value specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain a ligand residue name.\n" % (SelectionSpecParamName, ParamsOptionName))
        elif len(SelectionSpecList) != 1:
            MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain a single ligand residue name.\n" % (SelectionSpecParamValue, SelectionSpecParamName, ParamsOptionName))
        ResidueNames = SelectionSpecList
    elif re.match("^Residues$", SelectionParamValue, re.I):
        if SelectionSpecList is None:
            MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain a space delimited list of residue names.\n" % (SelectionSpecParamValue, SelectionSpecParamName, ParamsOptionName))
        ResidueNames = SelectionSpecList
    elif re.match("^Water$", SelectionParamValue, re.I):
        ResidueNames = ["HOH"]
        if SelectionSpecList is not None:
            ResidueNames.extend(SelectionSpecList)

    ParamsInfo["ResidueNames"] = [ResidueName.upper() for ResidueName in ResidueNames]
    ParamsInfo["CAlphaProteinStatus"] = True if re.match("^CAlphaProtein$", SelectionParamValue, re.I) else False

def ProcessOptionOpenMMForcefieldParameters(ParamsOptionName, ParamsOptionValue, ParamsDefaultInfo = None):
    """Process parameters for biopolymer, small molecule, and water forcefields and
    return a map containing processed parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    for forcefields.
    
    The supported parameter names along with their default and possible
    values are shown below:
    
        biopolymer, amber14-all.xml  [ Possible values: Any Valid value ]
        smallMolecule, OpenFF_2.2.0  [ Possible values: Any Valid value ]
        water, auto  [ Possible values: Any Valid value ]
    
    Possible biopolymer forcefield values:
    
        amber14-all.xml, amber99sb.xml, amber99sbildn.xml, amber03.xml,
        amber10.xml
        charmm36.xml, charmm_polar_2019.xml
        amoeba2018.xml
    
    Possible small molecule forcefield values:
    
        openff_2.2.0, openff_2.0.0, openff_1.3.1, openff_1.2.1, openff_1.1.1,
        smirnoff99frosst
        gaff-2.11, gaff-2.1, gaff-1.81, gaff-1.8, gaff-1.4
    
    The default water forcefield valus is dependent on the type of the
    biopolymer forcefield as shown below:
    
        Amber: amber14/tip3pfb.xml
        CHARMM: charmm36/water.xml or None for charmm_polar_2019.xml
        Amoeba: None (Explicit)
        
    Possible water forcefield values:
    
      amber14/tip3p.xml, amber14/tip3pfb.xml, amber14/spce.xml,
      amber14/tip4pew.xml, amber14/tip4pfb.xml,
      implicit/obc2.xml, implicit/GBn.xml, implicit/GBn2.xml
      charmm36/water.xml, charmm36/tip3p-pme-b.xml,
      charmm36/tip3p-pme-f.xml, charmm36/spce.xml,
      charmm36/tip4pew.xml, charmm36/tip4p2005.xml,
      charmm36/tip5p.xml, charmm36/tip5pew.xml,
      implicit/obc2.xml, implicit/GBn.xml, implicit/GBn2.xml
      amoeba2018_gk.xml (Implict water), None (Explicit water for amoeba)
        
    You may specify any valid forcefield name supported by OpenMM. No
    explicit validation is performed.

    Arguments:
        ParamsOptionName (str): Command line OpenMM forcefield option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"Biopolymer": "amber14-all.xml", "SmallMolecule": "openff-2.2.1", "Water": "auto", "Additional" : "None"}

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMForcefieldParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMForcefieldParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)

    return ParamsInfo

def _ProcessOptionOpenMMForcefieldParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue):
    """Process parameters with possible auto values and perform validation.
    """
    WaterForcefield = ParamsInfo["Water"]
    if re.match("^None$", WaterForcefield, re.I):
        WaterForcefield = None
        ParamsInfo["Water"] = WaterForcefield
    elif re.match("^auto$", WaterForcefield, re.I):
        BiopolymerForcefield = ParamsInfo["Biopolymer"]
        if re.search("amber", BiopolymerForcefield, re.I):
            WaterForcefield = "amber14/tip3pfb.xml"
        elif re.search("charmm", BiopolymerForcefield, re.I):
            if re.search("charmm_polar_2019", BiopolymerForcefield, re.I):
                WaterForcefield = None
            else:
                WaterForcefield = "charmm36/water.xml"
        elif re.search("amoeba", BiopolymerForcefield, re.I):
            # Explicit water...
            WaterForcefield = None
        else:
            WaterForcefield = None
        ParamsInfo["Water"] = WaterForcefield

    # Set status of implicit water forcefield...
    BiopolymerForcefield = ParamsInfo["Biopolymer"]
    ParamsInfo["ImplicitWater"] = True if _IsImplicitWaterForcefield(BiopolymerForcefield, WaterForcefield) else False

    # Process additional forcefields...
    ParamName = "Additional"
    ParamValue = ParamsInfo[ParamName]
    ParamValuesList = None
    if not re.match("^None$", ParamValue, re.I):
        ParamValuesList = ParamValue.split()
        if len(ParamValuesList) == 0:
            MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain a space delimited list of valid values.\n" % (ParamValue, ParamName, ParamsOptionName))
    ParamsInfo["AdditionalList"] = ParamValuesList

def _IsImplicitWaterForcefield(BiopolymerForcefield, WaterForcefield):
    """Check the nature of the water forcefield."""

    Status = False
    if WaterForcefield is None:
        if re.search("charmm_polar_2019", BiopolymerForcefield, re.I):
            Status = True
        else:
            Status = False
    else:
        if re.search("amber", BiopolymerForcefield, re.I):
            if re.search("implicit", WaterForcefield, re.I):
                Status = True
        elif re.search("charmm", BiopolymerForcefield, re.I):
            if re.search("implicit", WaterForcefield, re.I):
                Status = True
        elif re.search("amoeba", BiopolymerForcefield, re.I):
            Status = True

    return Status

def ProcessOptionOpenMMAnnealingParameters(ParamsOptionName, ParamsOptionValue, ParamsDefaultInfo = None):
    """Process parameters for annealing option and return a map containing processed
    parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    to setup platform.
    
    The supported parameter names along with their default values for
    different platforms are shown below:
        
        threads, 1  [ Possible value: >= 0 or auto.  The value of 'auto'
        Initial heating parameters:
            
        initialStart, 0.0  [ Units: kelvin ]
        initialEnd, 300.0  [ Units: kelvin ]
        initialChange, 5.0  [ Units: kelvin ]
        initialSteps, 5000
        
        initialEquilibrationSteps, 100000
        
        Heating and cooling cycle parameters:
        
        cycles, 1
        
        cycleStart, auto  [ Units: kelvin. The default value is set to
            initialEnd ]
        cycleEnd, 315.0  [ Units: kelvin ]
        cycleChange, 1.0  [ Units: kelvin ]
        cycleSteps, 1000
        
        cycleEquilibrationSteps, 100000
        
        Final equilibration parameters:
            
        finalEquilibrationSteps, 200000
            
    A brief description of parameters is provided below:
    
        Initial heating parameters:
        
        initialStart: Start temperature for initial heating.
        initialEnd: End temperature for initial heating.
        initialChange: Temperature change for increasing temperature
            during initial heating.
        initialSteps: Number of simulation steps after each
            heating step during initial heating
        
        initialEquilibrationSteps: Number of equilibration steps
            after the completion of initial heating.
        
        Heating and cooling cycles parameters:
        
        cycles: Number of annealing cycles to perform. Each cycle
            consists of a heating and a cooling phase. The heating phase
            consists of the following steps: Heat system from start to
            end temperature using change size and perform simulation for a
            number of steps after each increase in temperature; Perform
            equilibration after the completion of heating. The cooling
            phase is reverse of the heating phase and cools the system
            from end to start temperature.
        
        cycleStart: Start temperature for annealing cycle.
        cycleEnd: End temperature for annealing cycle.
        cycleChange: Temperature change for increasing or decreasing
            temperature during annealing cycle.
        cycleSteps: Number of simulation steps after each heating and
            cooling step during annealing cycle.
        
        cycleEquilibrationSteps: Number of equilibration steps
            after the completion of heating and cooling phase during a
            annealing cycle.
        
        Final equilibration parameters:
        
        finalEquilibrationSteps: Number of final equilibration
            steps after the completion of annealing cycles.

    Arguments:
        ParamsOptionName (str): Command line OpenMM annealing option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"InitialStart": 0.0, "InitialEnd": 300.0, "InitialChange": 5.0, "InitialSteps": 5000,  "InitialEquilibrationSteps": 100000, "Cycles": 1, "CycleStart": "auto", "CycleEnd": 315.0, "CycleChange": 1.0, "CycleSteps": 1000, "CycleEquilibrationSteps": 100000, "FinalEquilibrationSteps": 200000 }

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMAnnealingParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        if re.match("^(InitialStart|InitialEnd|CycleEnd)$", ParamName, re.I):
            if not MiscUtil.IsFloat(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
            Value = float(Value)
            if Value < 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^(InitialChange|CycleChange)$", ParamName, re.I):
            if not MiscUtil.IsFloat(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
            Value = float(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^CycleStart$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.IsFloat(Value):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
                Value = float(Value)
                if Value < 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
                ParamValue = Value
        elif re.match("^(InitialSteps|InitialEquilibrationSteps|Cycles|CycleSteps|CycleEquilibrationSteps|FinalEquilibrationSteps)$", ParamName, re.I):
            if not MiscUtil.IsInteger(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a integer.\n" % (Value, ParamName, ParamsOptionName))
            Value = int(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        else:
            ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMAnnealingParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)

    return ParamsInfo

def _ProcessOptionOpenMMAnnealingParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue):
    """Process parameters with possible auto values and perform validation.
    """

    ParamName = "CycleStart"
    ParamValue = "%s" % ParamsInfo[ParamName]
    if re.match("^auto$", ParamValue, re.I):
        ParamsInfo[ParamName] = ParamsInfo["InitialEnd"]

    if ParamsInfo["InitialStart"] >= ParamsInfo["InitialEnd"]:
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, initialStart, must be less than value, %s, specified for parameter name, initialEnd, using \"%s\" option.\n" % (ParamsInfo["InitialStart"], ParamsInfo["InitialEnd"], ParamsOptionName))

    if ParamsInfo["CycleStart"] >= ParamsInfo["CycleEnd"]:
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, cycleStart, must be less than value, %s, specified for parameter name, cycleEnd, using \"%s\" option.\n" % (ParamsInfo["CycleStart"], ParamsInfo["CycleEnd"], ParamsOptionName))

    if ParamsInfo["CycleStart"] != ParamsInfo["InitialEnd"]:
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, cycleStart, must be equal to value, %s, specified for parameter name, initialEndEnd, using \"%s\" option.\n" % (ParamsInfo["CycleStart"], ParamsInfo["InitialEnd"], ParamsOptionName))

    if ParamsInfo["InitialChange"] >= (ParamsInfo["InitialEnd"] - ParamsInfo["InitialStart"]):
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, initialChange, must be less than value, %s, corresponding to the difference between values specified for parameter names, initialStart and initialEnd, using \"%s\" option.\n" % (ParamsInfo["InitialChange"], (ParamsInfo["InitialEnd"] - ParamsInfo["InitialStart"]), ParamsOptionName))
        
    if ParamsInfo["CycleChange"] >= (ParamsInfo["CycleEnd"] - ParamsInfo["CycleStart"]):
        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, cycleChange, must be less than value, %s, corresponding to the difference between values specified for parameter names, cycleStart and cycleEnd, using \"%s\" option.\n" % (ParamsInfo["CycleChange"], (ParamsInfo["CycleEnd"] - ParamsInfo["CycleStart"]), ParamsOptionName))
        

def ProcessOptionOpenMMPlatformParameters(ParamsOptionName, ParamsOptionValue, ParamsDefaultInfo = None):
    """Process parameters for platform option and return a map containing processed
    parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    to setup platform.
    
    The supported parameter names along with their default values for
    different platforms are shown below:
    
        CPU:
        
        threads, 1  [ Possible value: >= 0 or auto.  The value of 'auto'
            or zero implies the use of all available CPUs for threading. ]
        
        CUDA:
        
        deviceIndex, auto  [ Possible values: 0, '0 1' etc. ]
        deterministicForces, auto [ Possible values: yes or no ]
        precision, single  [ Possible values: single, double, or mix ]
        tempDirectory, auto [ Possible value: DirName ]
        useBlockingSync, auto [ Possible values: yes or no ]
        useCpuPme, auto [ Possible values: yes or no ]
        
        OpenCL:
        
        deviceIndex, auto  [ Possible values: 0, '0 1' etc. ]
        openCLPlatformIndex, auto  [ Possible value: Number]
        precision, single  [ Possible values: single, double, or mix ]
        useCpuPme, auto [ Possible values: yes or no ]
    
    A brief description of parameters is provided below:
    
        CPU:
        
        threads: Number of threads to use for simulation.
        
        CUDA:
        
        deviceIndex: Space delimited list of device indices to use for
            calculations.
        deterministicForces: Generate reproducible results at the cost of a
            small decrease in performance.
        precision: Number precision to use for calculations.
        tempDirectory: Directory name for storing temporary files.
        useBlockingSync: Control run-time synchronization between CPU and
            GPU.
        useCpuPme: Use CPU-based PME implementation.
        
        OpenCL:
        
        deviceIndex: Space delimited list of device indices to use for
            simulation.
        openCLPlatformIndex: Platform index to use for calculations.
        precision: Number precision to use for calculations.
        useCpuPme: Use CPU-based PME implementation.

    Arguments:
        ParamsOptionName (str): Command line OpenMM platform option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"Name": "CPU", "Threads": "auto", "DeviceIndex": "auto", "DeterministicForces": "auto", "Precision": "single", "TempDirectory": "auto", "UseBlockingSync": "auto", "UseCpuPme": "auto", "OpenCLPlatformIndex": "auto"}

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMPlatformParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        if re.match("^Name$", ParamName, re.I):
            if not re.match("^(CPU|CUDA|OpenCL|Reference)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: CPU, GPU, OpenCL, or Reference" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^Threads$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not MiscUtil.IsInteger(Value):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be an integer.\n" % (Value, ParamName, ParamsOptionName))
                Value = int(Value)
                if Value < 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: >=0 \n" % (ParamValue, ParamName, ParamsOptionName))
                if Value > mp.cpu_count():
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is greater than number of CPUs, %s, returned by mp.cpu_count().\n" % (ParamValue, ParamName, ParamsOptionName, mp.cpu_count()))
                ParamValue = "%s" % Value
        elif re.match("^Precision$", ParamName, re.I):
            if not re.match("^(Single|Double|Mix)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: single, double or mix" % (Value, Name, ParamsOptionName))
            ParamValue = Value.lower()
        elif re.match("^(DeterministicForces|UseBlockingSync|UseCpuPme)$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not re.match("^(yes|no|true|false)$", Value, re.I):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: yes or no" % (Value, Name, ParamsOptionName))
                ParamValue = "true" if re.match("^(yes|true)$", Value, re.I) else "false"
        elif re.match("^(DeviceIndex|openCLPlatformIndex)$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                DeviceIndices = Value.split()
                if len(DeviceIndices) == 0:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain space delimited list of device indices.\n" % (Value, ParamName, ParamsOptionName))
                for DeviceIndex in DeviceIndices:
                    if not MiscUtil.IsInteger(DeviceIndex):
                        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be an integer.\n" % (DeviceIndex, ParamName, ParamsOptionName))
                ParamValue = ",".join(DeviceIndices)
        elif re.match("^TempDirectory$", ParamName, re.I):
            if not re.match("^auto$", Value, re.I):
                if not os.path.isdir(Value):
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. The specified directory doesn't exists." % (Value, Name, ParamsOptionName))
            ParamValue = Value
        else:
            ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMPlatformParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)

    return ParamsInfo

def _ProcessOptionOpenMMPlatformParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue):
    """Process parameters with possible auto values and perform validation.
    """

    ParamValueMap = {"cpu": "CPU", "cuda": "CUDA", "opencl": "OpenCL", "reference": "Reference"}
    ParamName = "Name"
    ParamValue = ParamsInfo[ParamName].lower()
    if ParamValue in ParamValueMap:
        ParamsInfo[ParamName] = ParamValueMap[ParamValue]

    ParamsInfo["Precision"] = ParamsInfo["Precision"].lower()

    # Set "auto" values to None and treat all other values as strings...
    for ParamName in ParamsInfo:
        ParamValue = "%s" % ParamsInfo[ParamName]
        if re.match("^auto$", ParamValue, re.I):
            ParamsInfo[ParamName] = None
        else:
            ParamsInfo[ParamName] = ParamValue

def ProcessOptionOpenMMWaterBoxParameters(ParamsOptionName, ParamsOptionValue, ParamsDefaultInfo = None):
    """Process parameters for adding a water box option and return a map containing
    processed parameter names and values.
    
    ParamsOptionValue is a comma delimited list of parameter name and value pairs
    for adding a water box.
    
    The supported parameter names along with their default and possible
    values are shown below:
    
        model, tip3p [ Possible values: tip3p, spce, tip4pew, tip5p or swm4ndp ]
        mode, Padding [ Possible values: Size or Padding ]
        size, None [ Possible values: xsize ysize zsize ]
        padding, 1.0
        shape, cube [ Possible values: cube, dodecahedron, or octahedron ]
        ionPositive, Na+ [ Possible values: Li+, Na+, K+, Rb+, or Cs+ ]
        ionNegative, Cl- [ Possible values: Cl-, Br-, F-, or I- ]
        ionicStrength, 0.0
    
    A brief description of parameters is provided below:
    
    model: Water model to use for adding water box.
    
    mode: Specify the size of the waterbox explicitly or calculate it automatically
    for a macromolecule along with adding padding around macromolecule.
    Possible values: Size or Padding.
    
    size: A space delimited triplet of values corresponding to water size in
    nanometers. It must be specified during 'Size' value of 'mode' parameter.
    
    padding: Padding around macromolecule in nanometers for filling box with
    water. It must be specified during 'Padding' value of 'mode' parameter.
    
    ionPositive: Type of positive ion to add during the addition of a water box.
    
    ionNegative: Type of negative ion to add during the addition of a water box.
    
    ionicStrength: Total concentration (molar) of both positive and negative ions
    to add excluding he ions added to neutralize the system during the addition
    of a water box.

    Arguments:
        ParamsOptionName (str): Command line OpenMM water box option name.
        ParamsOptionValues (str): Space delimited list of parameter name and value pairs.
        ParamsDefaultInfo (dict): Default values to override for selected parameters.

    Returns:
        dictionary: Processed parameter name and value pairs.

    """

    ParamsInfo = {"Model": "tip3p", "Mode": "Padding", "Size": None, "Padding": 1.0, "Shape": "cube", "IonPositive": "Na+", "IonNegative": "Cl-", "IonicStrength": 0.0}

    (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords) = _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo)

    if re.match("^auto$", ParamsOptionValue, re.I):
        _ProcessOptionOpenMMWaterBoxParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)
        return ParamsInfo

    for Index in range(0, len(ParamsOptionValueWords), 2):
        Name = ParamsOptionValueWords[Index].strip()
        Value = ParamsOptionValueWords[Index + 1].strip()

        ParamName = CanonicalParamNamesMap[Name.lower()]
        ParamValue = Value

        if re.match("^Model$", ParamName, re.I):
            if not re.match("^(tip3p|spce|tip4pew|tip5p|swm4ndp)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: tip3p, spce, tip4pew, tip5p, or swm4ndp" % (Value, Name, ParamsOptionName))
            ParamValue = Value.lower()
        elif re.match("^Mode$", ParamName, re.I):
            if not re.match("^(Padding|Size)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: Padding or Size" % (Value, Name, ParamsOptionName))
            ParamValue = Value
        elif re.match("^Size$", ParamName, re.I):
            if Value is not None and not re.match("^None$", Value, re.I):
                SizeValues = Value.split()
                if len(SizeValues) != 3:
                    MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. It must contain 3 float values separated by spaces.\n" % (Value, ParamName, ParamsOptionName))

                SizeValueList = []
                for SizeValue in SizeValues:
                    if not MiscUtil.IsFloat(SizeValue):
                        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (SizeValue, ParamName, ParamsOptionName))
                    SizeValue = float(SizeValue)
                    if SizeValue <= 0:
                        MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (SizeValue, ParamName, ParamsOptionName))
                    SizeValueList.append(SizeValue)

                # Set size values...
                ParamsInfo["SizeList"] = SizeValueList

            ParamValue = Value
        elif re.match("^Padding$", ParamName, re.I):
            if not MiscUtil.IsFloat(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
            Value = float(Value)
            if Value <= 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: > 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        elif re.match("^Shape$", ParamName, re.I):
            if not re.match("^(cube|dodecahedron|octahedron)$", Value, re.I):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: cube, dodecahedron, or octahedron" % (Value, Name, ParamsOptionName))
            ParamValue = Value.lower()
        elif re.match("^IonPositive$", ParamName, re.I):
            ValidValues = "Li+ Na+ K+ Rb+ Cs+"
            EscapedValidValuesPattern = "Li\+|Na\+|K\+|Rb\+|Cs\+"
            if not re.match("^(%s)$" % EscapedValidValuesPattern, Value):
                MiscUtil.PrintError("The value specified, %s, for parameter name, %s, using  \"%s\" option is not a valid.  Supported value(s): %s" % (ParamValue, ParamName, ParamsOptionName, ValidValues))
            ParamValue = Value
        elif re.match("^IonNegative$", ParamName, re.I):
            ValidValues = "F- Cl- Br- I-"
            ValidValuesPattern = "F-|Cl-|Br-|I-"
            if not re.match("^(%s)$" % ValidValuesPattern, Value):
                MiscUtil.PrintError("The value specified, %s, for parameter name, %s, using  \"%s\" option is not a valid.  Supported value(s): %s" % (ParamValue, ParamName, ParamsOptionName, ValidValues))
            ParamValue = Value
        elif re.match("^IonicStrength$", ParamName, re.I):
            if not MiscUtil.IsFloat(Value):
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option must be a float.\n" % (Value, ParamName, ParamsOptionName))
            Value = float(Value)
            if Value < 0:
                MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: >= 0\n" % (ParamValue, ParamName, ParamsOptionName))
            ParamValue = Value
        else:
            ParamValue = Value

        # Set value...
        ParamsInfo[ParamName] = ParamValue

    # Handle parameters with possible auto values...
    _ProcessOptionOpenMMWaterBoxParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue)

    return ParamsInfo

def _ProcessOptionOpenMMWaterBoxParameters(ParamsInfo, ParamsOptionName, ParamsOptionValue):
    """Process parameters with possible auto values and perform validation.
    """

    ParamsInfo["ModeSize"] = True if re.match("^Size$", ParamsInfo["Mode"], re.I) else False
    ParamsInfo["ModePadding"] = True if re.match("^Padding$", ParamsInfo["Mode"], re.I) else False

    if ParamsInfo["ModeSize"]:
        ParamName = "Size"
        ParamValue = ParamsInfo[ParamName]
        if ParamValue  is None:
            MiscUtil.PrintError("The parameter value, %s, specified for parameter name, %s, using \"%s\" option is not a valid value. Supported values: x y z\n" % (ParamValue, ParamName, ParamsOptionName))
    else:
         ParamsInfo["SizeList"] = None 

def _ValidateAndCanonicalizeParameterNames(ParamsOptionName, ParamsOptionValue, ParamsInfo, ParamsDefaultInfo):
    """Validate and canonicalize parameter names."""

    # Setup a canonical paramater names...
    ValidParamNames = []
    CanonicalParamNamesMap = {}
    for ParamName in sorted(ParamsInfo):
        ValidParamNames.append(ParamName)
        CanonicalParamNamesMap[ParamName.lower()] = ParamName

    # Update default values...
    if ParamsDefaultInfo is not None:
        for ParamName in ParamsDefaultInfo:
            if ParamName not in ParamsInfo:
                MiscUtil.PrintError("The default parameter name, %s, specified using \"%s\" option is not a valid name. Supported parameter names: %s" % (ParamName, ParamsDefaultInfo, " ".join(ValidParamNames)))
            ParamsInfo[ParamName] = ParamsDefaultInfo[ParamName]

    ParamsOptionValue = ParamsOptionValue.strip()
    if not ParamsOptionValue:
        MiscUtil.PrintError("No valid parameter name and value pairs specified using \"%s\" option" % ParamsOptionName)

    ParamsOptionValueWords = None
    if not re.match("^auto$", ParamsOptionValue, re.I):
        ParamsOptionValueWords = ParamsOptionValue.split(",")
        if len(ParamsOptionValueWords) % 2:
            MiscUtil.PrintError("The number of comma delimited paramater names and values, %d, specified using \"%s\" option must be an even number." % (len(ParamsOptionValueWords), ParamsOptionName))

    if ParamsOptionValueWords is not None:
        for Index in range(0, len(ParamsOptionValueWords), 2):
            Name = ParamsOptionValueWords[Index].strip()
            CanonicalName = Name.lower()
            if  not CanonicalName in CanonicalParamNamesMap:
                MiscUtil.PrintError("The parameter name, %s, specified using \"%s\" is not a valid name. Supported parameter names: %s" % (Name, ParamsOptionName, " ".join(ValidParamNames)))
        
    return (ValidParamNames, CanonicalParamNamesMap, ParamsOptionValue, ParamsOptionValueWords)
